
using System;
using System.Text;
using System.IO;
using System.Collections.Generic;
using System.Reflection;
using System.Linq;
using System.Diagnostics;
using VisTool;

namespace VisTool {

public partial class Parser {
    public static PathInfo pi; // Keeps track of current state of the formula.
    // pi is static to allow Intellisense to follow the state. This means that pi must be stacked 
    // and unstacked in some places, e.g. when CompileRows calls the Parser recursively.
    static bool cutToken = false; // True if EOL tokens are to be replaced with Operator tokens and marked cutToken.
    // Intellisense depends on this. SL 10-12-16

    public Parser(Template tpl, TextReader r, TplProperty p, bool cut) { // Used with memory texts (strings)
        cutToken = cut; // When reading a memory text, the last token may be cut (terminated with EOF)
        tplProp = p;
        propName = p.Name; // For error reporting
        tplName = tpl.Name; // For error reporting
        this.tpl = tpl;
        //frmName = Template.ParentTemplate==null ? Template.Name : ??
        if (p.startChain != null)
            la = tokens = pt = p.startChain; // The formula has been scanned already - during OpenVisFile
        else {
            rd = r; 
            la = tokens = pt = NextToken(); // There is an la-token. No token read.
            p.startChain = la; // Start the chain of tokens
        }
        pi = new PathInfo(tpl, PathState.Init); 
    }

    public void Formula(TplProperty tplProp) { // Formula = [ EventHandler | Expressions ]  So a formula may be empty.
        // The Rows formula is handled by CompileRows and in case of crossrefs by PrepareRowTpl.
        // When non-row formulas are compiled, the rows formula has been compiled already.
        // Debug.Print(Template.Name + ": " + propName + ":  "+ p.Formula);
        pi.Tpl = tpl; pi.RowTpl = tpl.RowTpl;
        Expression expr = null;
        bool init = false;
        tplProp.expr = null;
        // Debug.Print(Template.Name + ": " + propName + ":  "+ p.Formula);
        IsStatement = false;

        try {
            if (tplProp is TplPropEvent) {
                pi.state = PathState.ExpectStatement; // Set state to help Intellisense.
                if (La.EOF) {
                    if (la.comment)
                        pi.state = PathState.FreeText; // The token may hold comments, unterminated strings, etc.
                    return; // Empty formula or partial keyword
                }
                EventHandler(tplProp);
            } else if (tplProp == tpl.propRows)
                rowsFormula(tplProp); // Access this way too. Intellisense doesn't know.
            else if (La.val.ToUpper() == "FIND")
                findFormula(tplProp);
            else { // Normal property
                // Expressions = [ [Init] expression { , [Init] expression} ] EOL
                // If the formula is a list with more than one element, p.expr becomes an expression list.Otherwise an ordinary expression.

                if (la.EOF) { // Empty formula?
                    if (la.comment)
                        pi.state = PathState.FreeText; // The token may hold comments, unterminated strings, etc.
                    return; // Empty formula or partial keyword
                }
                if (La.val.ToUpper() == "INIT") { Get(); init = true; }
                pi.state = PathState.Not; 
                pi.brackets.Push(",");
                defaultExpr(ref expr); // Non-empty expression required
                if (la.kind == TkKind.EOL) { // One element only. Ordinary expression. 
                    tplProp.expr = expr;
                } else {
                    Expect(",");
                    if (expr != null) expr.Init = init;
                    tplProp.expr = new ExpressionList(); // The ExpressionList has init=false, but the elements may have init = true.
                    (tplProp.expr as ExpressionList).Add(expr);
                    while (true) {
                        pi.state = PathState.Init; 
                        init = false;
                        if (La.kind == TkKind.EOL) { // Empty formula.
                            error("Expects an operand (field, property, 123, \"abc\") or a unary operator (Init, -, Not)", la);
                        }
                        if (la.val.ToUpper() == "INIT") { Get(); init = true; }
                        pi.state = PathState.Not; 
                        defaultExpr(ref expr);
                        if (expr != null) expr.Init = init;
                        (tplProp.expr as ExpressionList).Add(expr);
                        if (La.val != ",") break;
                        Get();
                    };
                    init = false;
                }
                if (tplProp.expr != null) { // SL 02-07-12
                    tplProp.expr.Init = init;
                }
            }
            if (formulaError) throw new LoggedException(); // There was an error on the way
            Expect(TkKind.EOL);
        }
        catch (LoggedException) { // The error was logged earlier or there was a cutToken without an error.
                tplProp.expr = null; // Like an empty formula, but only when not cutToken.
        }
        catch (Exception e) { // Unexpected exception.
            if (e is CutException) throw e;
            ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, tpl, tplProp, e.Message, token, 0);
            tplProp.expr = null; // Like an empty formula
        }
        if (la.comment) 
            pi.state = PathState.FreeText; // Last token ended with a comment but no \n.
        pi.la = La; // Allow Intellisense to see the last unrecognized token.
    }

    void expression(ref Expression e) { // expression = defaultExpr queryTail.  Note: a join is a path (which is a simpleExpr)
        defaultExpr(ref e); queryTail(ref e);
        // At present (18-06-13), queries are only allowed in Rows, but in principle they might be allowed in design properties and as parameters.
    }

    public void rowsFormula(TplProperty tplProp) {  // Compile the Rows formula.
        // rows = ([Map.] table { join relation} queryTail |    tpl.RowsExpr = RecordSetExpr; tpl.RowTpl = tpl
        //             template |                                                 tpl.RowsExpr = TemplateRefExpr; tpl.RowTpl = template
        //             Create int |                                               tpl.RowsExpr = ??; tpl.RowTpl = tpl
        //             null                                                           tpl.RowsExpr = null; tpl.RowTpl = tpl.
        pi.Tpl = tpl;
        Expression expr = null;
        tpl.RowsExpr = null; tplProp.expr = null; tpl.RowTpl = tpl; // Default values.
        if (tpl.Visited) {
            error("Circular reference in Rows", null);
            tpl.RowTpl.RowsCompiled = true; return;
        }
        pi.state = PathState.Rows; // Set state to help Intellisense when the Rows formula starts with a cut ident.

        if (La.EOF) { // Empty formula or partial keyword
            if (la.comment)
                pi.state = PathState.FreeText; // The token may hold comments, unterminated strings, etc.
            tpl.RowsCompiled = true; return;
        }
        tpl.Visited = true;
        try {
            if (La.val.ToUpper() == "CREATE") {
                Expression count = null;
                defaultExpr(ref count); // May cause crossrefs.
                tplProp.expr = tpl.RowsExpr = new CreateRows(count, tpl);
            } else { // Template (shared rows) or real query.
                pi.state = PathState.Rows; 
                Path(out expr);
                TemplateRefExpr tplRef = expr as TemplateRefExpr;
                FrmCtrlExpr frmRef = expr as FrmCtrlExpr;
                if (tplRef != null) {
                    Template target = tplRef.targetTpl; // Shared rows. target points to it. Cannot be foreign template.
                    if (!target.RowsCompiled) {
                        CompileTarget(target);
                    }
                    if (!(target.propRows.expr is RecordSetExpr || target.propRows.expr is CreateRows) )
                        error("Doesn't refer to a template with rows", token); 
                    else
                        tpl.RowTpl = target;
                } else if (frmRef != null) {
                    tpl.RowTpl = tpl.FormTpl;
                } else {// Query formula.
                    queryTail(ref expr);
                    RecordSetExpr rs = (RecordSetExpr)expr;
                    if (rs != null && rs.exprGroupby.Count == 0) { // Cannot assign to Group By queries.
                        // Add key fields to support assignment to the fields at run time.
                        foreach (Join join in rs.joinList)
                            rs.AddKeyFields(join);
                    }
                    tplProp.expr = tpl.RowsExpr = rs;
                    if (formulaError) throw new LoggedException(); // There was an error on the way.
                    Expect(TkKind.EOL);
                }
            }
        }
        catch (CutException) {
            tpl.Visited = false; throw new CutException();
        }
        catch (LoggedException) { // The error was logged earlier.
        }
        tpl.Visited = false;
        tpl.RowsCompiled = true;
        //Debug.Print("rowsFormula: " + tpl.Name);
    }

    void findFormula(TplProperty tplProp) {  // Compile the Find formula and update the target template's targetCount.
        // findFormula = Find targetTpl [ On targetKeyExpr = sourceKeyExpr { And targetKeyExpr = sourceKeyExpr} ] [ Order By sourceExpr {, sourceExpr}]
        // TargetKeyExpr and sourceKeyExpr may be interchanged.
        Template target;
        Expression expr1=null, expr2=null;
        List<Expression> toKeyExpr = new List<Expression>(4);
        List<Expression> fromKeyExpr = new List<Expression>(4);
        List<Expression> orderExpr = null; // Null when no Order By
        List<int> asc = new List<int>(4);

        Get(); pi.state = PathState.Find; // For Intellisense.
        if (!tpl.FormTpl.Templates.TryGetValue(La.val, out target) || target.Ignore) { error("Expects a template name", la); return; }
        Get();
        FindExpr findExpr = new FindExpr(tplProp, target, toKeyExpr, fromKeyExpr, orderExpr, asc, tplProp.startChain);
        tplProp.expr = findExpr; // Make expr already here. It will be meaningfull even if the rest fails.

        try {
            pi.state = PathState.FindTpl; 
            if (La.val.ToUpper() == "ON") {
                Get();
                pi.state = PathState.Not; 
                while (true) {
                    pi.RowTpl = target.RowTpl; pi.Tpl = target; pi.block = tpl; pi.state = PathState.Not; 
                    Add(out expr1); Expect("=");
                    pi.RowTpl = tpl.RowTpl; pi.Tpl = tpl; pi.block = target; pi.state = PathState.Not; 
                    Add(out expr2);
                    pi.block = null;
                    toKeyExpr.Add(expr1); fromKeyExpr.Add(expr2);
                    if (La.val.ToUpper() != "AND") break;
                    Get();
                }
                pi.state = PathState.FindOn; 
                findExpr.toKeyExpr = toKeyExpr; findExpr.fromKeyExpr = fromKeyExpr;
            }

            if (La.val.ToUpper() == "ORDER") {
                Get();
                pi.state = PathState.ExpectSymbol; 
                Expect("By");
                orderExpr = new List<Expression>(4);
                while (true) {
                    int ascItem = 1;
                    pi.RowTpl = tpl; pi.state = PathState.Not;
                    defaultExpr(ref expr1); orderExpr.Add(expr1);
                    PathState old = pi.state;
                    pi.state = PathState.AscDesc;  // To support Intellisense.
                    if (La.val.ToUpper() == "DESC") { ascItem = -1; Get(); } else if (La.val.ToUpper() == "ASC") Get();
                    pi.state = old; // To support Intellisense.
                    asc.Add(ascItem);
                    if (La.val != ",") break;
                }
            }
            findExpr.orderBy = orderExpr; findExpr.asc = asc;
        }
        catch (Exception) { // Some expressions don't return a value at exception. Just ignore it..
        }

        int index = target.findProperties.IndexOf(tplProp);
        if (index < 0) target.findProperties.Add(tplProp);
    }

    private void EventHandler( TplProperty p ) { // eventHandler = statementList EOL
        // p.expr becomes an expression list, also when there is only one statement.
        // Errors are caught in Formula
        IsStatement = true; // Allows statement-only methods to be called.
        forStack.Clear();
        p.expr = null;
        StatementList(ref p.expr, p);
        //pi.state = PathState.EndStatement; // Set state to help Intellisense.
        Expect(TkKind.EOL);
    }

    private void StatementList(ref Expression stmt, TplProperty tplProp) { 
        // statementList = statement { ; statement } | "{" statement "}" statement
        // p.expr becomes an expression list, also when there is only one statement.
        // Errors are caught in Formula
        Expression expr = null;
        Statement(ref expr, tplProp);
        stmt = new StatementList(tplProp);
        if (expr != null) (stmt as StatementList).Add(expr);
        if (La.val == "}") return;
        while (la.val == ";" || token.val == "}") { // Previous statement terminating with } is ok too. 
            if (la.val == ";") Get();
            pi.state = PathState.ExpectStatement;
            if (La.kind == TkKind.EOL) break;
            Statement(ref expr, tplProp);
            if (expr != null) (stmt as StatementList).Add(expr); // Empty statements allowed.
            if (La.val == "}") return;
        }
    }

    public static List<string> forStack = new List<string>(); // Names of loop variables in nested For loops.
    void Statement(ref Expression stmt, TplProperty tplProp) { // statement = If (ifExpr) thenStatements [ Else elseStatements ] | path [ '=' expression ] 
        stmt = null;
        pi.state = PathState.ExpectStatement;  // Set state to help Intellisense.
        if (La.kind == TkKind.EOL || la.val == ";" || la.val == "}") // && pi.brackets.Count > 0 && pi.brackets.Peek() == "}") 
            return; // Empty statement. 
        if (la.val.ToUpper() == "RETURN") {
            stmt = new ReturnExpr(la);
            pi.state = PathState.EndStatement;
            Get(); return;
        }
        pi.brackets.Push(";");

        if (la.val == "{") { // CompositeStatement = "{" StatementList "}"   
            Get(); pi.brackets.Push("; }");
            StatementList(ref stmt, tplProp);
            Expect("}"); pi.brackets.Pop();
        } else

        if (la.val.ToUpper() == "FOR") { // For statement = For ( ident = initExpr ; conditionExpr ; ident = stepExpr )  [ "{ StatementList"}" | Statement ]  
            pi.state = PathState.ExpectSymbol; 
            Get(); Expect("("); pi.brackets.Push(";");
            Token loopVar;
            Expression initExpr = null;
            Expression conditionExpr = null;
            Expression stepExpr = null;
            Expression loopStatement = null;

            Expect(TkKind.Ident); loopVar = token;
            for (int i = 0; i < forStack.Count; i++)
                if (token.val.ToUpper() == forStack[i].ToUpper()) { error("Loop variable " + token.val + " is used already", token); throw new LoggedException(); }
            //if (forStack.Count >= 10) { error("Cannot handle more than 10 nested For loops", token); throw new LoggedException(); }
            Expect("=");
            pi.state = PathState.Unknown; // An expression, maybe start with Not.
            conditionalExpr(out initExpr);
            Expect(";");
            forStack.Add(loopVar.val); // Add the loop variable here in order that Intellisense doesn't suggest it until now.
            pi.state = PathState.Not; // An expression, maybe start with Not.
            defaultExpr(ref conditionExpr);
            Expect(";");
            pi.state = PathState.ExpectSymbol; 
            Expect(loopVar.val);
            Expect("="); pi.brackets.Pop(); pi.brackets.Push(")");
            pi.state = PathState.Not; // An expression, maybe start with Not.
            defaultExpr(ref stepExpr);
            Expect(")"); pi.brackets.Pop();

            Statement(ref loopStatement, tplProp);
            if ( !formulaError) forStack.RemoveAt(forStack.Count - 1); // Unstack. Keep stack for Intellisense.
            stmt = new ForLoop(initExpr, conditionExpr, stepExpr, loopStatement, forStack.Count, loopVar);
        } else

        if (la.val.ToUpper() == "IF") { // If statement = If (condition) [ "{ StatementList"}" | Statement ] [ Else [ "{ StatementList"}" | Statement ] ] 
            pi.state = PathState.ExpectSymbol; 
            Get(); Expect("("); pi.brackets.Push(")");
            Expression condition = null;
            Expression thenStatement = null;
            Expression elseStatement = null;
            pi.state = PathState.Not; 
            defaultExpr(ref condition);
            Expect(")"); pi.brackets.Pop();

            Statement(ref thenStatement, tplProp);
            pi.state = PathState.Else; 
            if (La.val.ToUpper() == "ELSE") {
                Get();
                pi.state = PathState.ExpectStatement; 
                Statement(ref elseStatement, tplProp);
            }
            stmt = new IfThen(condition, thenStatement, elseStatement, la);

        } else if (La.kind != TkKind.EOL) { //Method call or Assignment statement
            Expression lExp = null;
            Expression rExp = null;
            pi.state = PathState.Unknown; 
            Path(out lExp); // Reads function parameters too
            bool assignAble = lExp != null && lExp.GetType().GetMethod("Assign").DeclaringType != typeof(Expression);
            if (assignAble) {
                pi.state = PathState.ExpectSymbol; 
                Expect("=");
                Token tokenEqual = token;
                pi.state = PathState.Not; 
                conditionalExpr(out rExp); // Default not allowed
                HandleFieldAssignment(lExp, tokenEqual); // Adds key fields to the rows query when assigning to a field. Also checks for foreign form.
                stmt = new Assignment(tplProp, lExp, rExp, tokenEqual);
            } else if (lExp is FunctionCallExpr) {
                stmt = lExp;
            } else if (La.val == "=") { // Assignment statement
                ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, tplProp.tpl, tplProp.Name, "Cannot assign to this", token, 0);
                    throw new LoggedException();
            } else {
                ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, tplProp.tpl, tplProp.Name, "Unexpected symbol", token, 0);
                throw new LoggedException();
            }
        }
        pi.state = PathState.EndStatement; 

        if (la.val.ToUpper() == "DEFAULT") { // Allow continuation without a next token.
            Expression defaultExpr = null;
            Token defaultToken = la;
            defaultToken = la;
            Get(); Statement(ref defaultExpr, tplProp);
            stmt = new DefaultExpr(stmt, defaultExpr, defaultToken);
        }
        pi.state = PathState.EndStatement;
        pi.brackets.Pop();
    }

    bool IsLoopVariable(ref Expression e) {
        for (int level = 0; level < forStack.Count; level++ )
            if (forStack[level].ToUpper() == La.val.ToUpper()) {
                Get();
                e = new LoopVariable(level);
                return true;
            }
        return false;
    }

    void Path(out Expression result) {
        // Path = ( ( Me | Main | Form ) [index] | ( Forms | Templates | System | Map ) dot ident | [dot] ident )  
        //            { dot ident [index] } [ functionParameters ]
        // Handles an initial cutToken by means of the initial pi values.
        // Exceptions are caught by Formula, etc. This gives Intellisense access to the last "pi.la" across all errors.
        pi.dot = pi.ident = null; pi.leftNode = result = null;
        bool oldInTailExpr = pi.inTailExpr;
        Template oldTpl = pi.Tpl;
        string oldForeignForm = pi.foreignForm;
        //Debug.Print("Restored pi.Tpl: " + pi.Tpl.Name);


        if (IsLoopVariable(ref result)) { pi.state = PathState.Value; return; }
        if (pi.state != PathState.Rows) { pi.state = PathState.Unknown; }

        startPath = La; // Prepare for path being control stuff in an sqlTail. Collect the tokens to insert in the sql.

        try {
            //object parameters = new List<Expression>();
            if (la.kind == TkKind.Ident) {
                pi.ident = la; pi.dot = null;
                switch (la.val.ToUpper()) {
                case "ME": // Walks to the current control
                    Get(); pi.state = PathState.Control; pi.leftNode = new MeExpr(token);
                    // PrepareT_Rows() done initially;
                    subscript(); goto nextStep; // Optional index
                case "MAIN": // Walks to the main control (shared rows). (In earlier versions called Parent.
                    Get(); pi.state = PathState.Control; pi.leftNode = new MainExpr(token);
                    pi.Tpl = pi.RowTpl= tpl.RowTpl; 
                    subscript(); goto nextStep; // Optional index
                case "FORM": // Walks to the current form control
                    Get();
                    pi.state = PathState.Control; // PathState.Rows ? PathState.ParentTemplate : PathState.Control;
                    pi.leftNode = new FrmCtrlExpr(token);
                    pi.Tpl = tpl.FormTpl; pi.RowTpl = pi.Tpl.RowTpl;
                    subscript(); goto nextStep; // Optional index
                case "FORMS": // Walks to the collection of form controls (open forms)
                    Get(); pi.state = PathState.Forms; break;
                case "TEMPLATES": // Walks to the collection of form templates
                    Get(); pi.state = PathState.Templates; break;
                case "SYSTEM": // Walks to the collection of System functions
                    Get(); pi.state = PathState.System; break;
                case "MAP": // Walks to the collection of tables
                    if (tplProp.Name.ToUpper() != "ROWS")
                        error("Use the Rows property to access tables", token, true);
                    Get(); pi.state = PathState.Map; break;
                default: goto ident; // Non-prefix ident
                } // End switch
                if (!dotOps()) { error("Dot-operator expected", la); throw new LoggedException(); }
            } else if (checkDotOps()) {
                pi.state = PathState.Control; 
                goto nextStep;
            } else return;

ident:
            Expect(TkKind.Ident); // Any ident allowed after the prefix
            pi.ident = token;
            if (!CheckStep()) { error(ExpectText(), token); throw new LoggedException(); }
            subscript(); // optional [index]
nextStep:
            if (pi.inTailExpr && ! (pi.state == PathState.Field || pi.state == PathState.Aggregate ||
                (pi.leftNode is FunctionCallExpr) && (pi.leftNode as FunctionCallExpr).SQLsupport) ) {
                // Component stuff in sql tail. Generate sql snippet until here.                
                Expression t = new Constant<string>(MakeSnippet(startSnippet, startPath));
                // Make all tokens from start of sql snippet to start of path an sql string.
                if (tail == null) tail = t; else tail = new SQLConcatExpr(tail, t);
                // Compile the rest of the path in normal mode, including possible parameters and indexes.
                pi.inTailExpr = false;
                // The current template, pi.Tpl, may be outside the row template hierarchy being compiled.
                pi.RowTpl = pi.Tpl;
                //while (pi.RowTpl != null && pi.RowTpl.RowsExpr == null) pi.RowTpl = pi.RowTpl.ParentTemplate;
            }

            while (dotOps()) {
                Expect(TkKind.Ident); // Any ident allowed after a dot operator
                pi.ident = token; // CheckStep relies on pi.dot being the dot before the identifier.
                if (!CheckStep()) { error(ExpectText(), token); throw new LoggedException(); }
                subscript(); // optional [index]
            }
            if (pi.state == PathState.Record) pi.leftNode = pi.tempRecordRefExpr;
            result = pi.leftNode;
            if (pi.state == PathState.Aggregate) return; // Aggregate parameters read later
            if (oldInTailExpr && !pi.inTailExpr) { // The path was control stuff inside a query tail
                if (tail == null) tail = new SQLcontrolProperty(result);
                else tail = new SQLConcatExpr(tail, new SQLcontrolProperty(result));
                startSnippet = La;
                //pi.Tpl = tpl; PrepareRowTpl();  // Tpl and RowTpl may be changed during the path. Set them back to the current template
            }
        }
        catch (LoggedException) { // The error was logged earlier. 
            // Skip path and continue parsing to find more errors and reach any cutToken to support Intellisense.
            formulaError = true;
            while (dotOps()) {
                if (La.kind == TkKind.Ident) Get(); else break;
            }
            result = pi.leftNode;
        }
        pi.Tpl = oldTpl; pi.RowTpl = oldTpl.RowTpl;
        pi.foreignForm = oldForeignForm;
        pi.inTailExpr = oldInTailExpr;
    } // End: path

    void subscript() { // indexBracket = [ '[' expression ']' ]. Subscript is used in many states, e.g. tpl[i], bundle[i], listProp[i]. 
        // To support indirect property reference and foreign tpl reference, resolution is done at run time.
        if (la.val == "[") {
            if (pi.state != PathState.Control && pi.state != PathState.Property) {
                error("Only components and properties can have a subscript", token); throw new LoggedException(); 
            }
            pi.brackets.Push("]"); // Stack expected closing character.
            Get();
            PathInfo old = pi; pi = new PathInfo(tpl, PathState.ExpectPrimary);
            pi.aggregateName = old.aggregateName; // // Fields are allowed depending on whether the whole expression is in an aggregate call.
            pi.leftNode = old.leftNode; // Index and Last need leftNode to evaluate relative to it.
            Expression eIndex = null;
            defaultExpr(ref eIndex);
            pi = old; // Unstack
            Expect("]");
            pi.brackets.Pop(); // Unstack only if there is a matching parenthesis
            pi.leftNode = new Subscript(pi.leftNode, eIndex, token);
            if (pi.state != PathState.Control && pi.state != PathState.Property) { // Control and Property have the same state after subscript.
                pi.state = PathState.Value; 
            }
        }
    }

    void functionCall(MethodInfo targetMethod, bool SQLsupport) { // Fnc ( [ expression { , expression } ] 
        // Reads the parameter list, checks that it matches the signature of Fnc, and adds a FunctionCall expression to the leftNode path.
        List<Expression> parameters = new List<Expression>(8);
        Token tokenFunction = token;
        PathInfo old = pi; pi = new PathInfo(tpl, PathState.Not);
        Expect("(");
        if (targetMethod.GetParameters().Length > 0) { // One or more parameters required.
            pi.aggregateName = old.aggregateName; // Fields are allowed depending on whether the whole expression is in an aggregate call.

            if (SQLsupport && old.inTailExpr)
                pi.inTailExpr = true; // Allow field refs in the parameters
            pi.brackets.Push(", )"); // Stack expected closing character.
            Expression e2 = null;
            if (!(La.val == ")")) {
                pi.state = PathState.Not;
                defaultExpr(ref e2);
                parameters.Add(e2);
                while (La.val == ",") {
                    Get();
                    pi.state = PathState.Not;
                    defaultExpr(ref e2);
                    parameters.Add(e2);
                }
            }
            pi.brackets.Pop(); // Unstack only if there is a matching parenthesis
        } else { // Zero parameters required.
            pi.state = PathState.ExpectSymbol; 
        }
        Expect(")");
        pi = old;
        pi.state = PathState.Value; 
        if (targetMethod.ReturnType == typeof(VisComponent))
            pi.state = PathState.Control; // At present the only non-value type returned by a function.

        // Parameter list mismatch is reported with template and property name through .NET exceptions caught by Formula, etc.
        ParameterInfo[] formalParams = targetMethod.GetParameters();
        List<Expression> actualParamList = parameters as List<Expression>;
        int actualLength = actualParamList.Count;
        for (int i = 0; i < formalParams.Length; i++) {
            Type formalType = formalParams[i].ParameterType;
            // Read the formal parameters one by one and add a type-conversion node in special cases.
            if (formalType == typeof(object[]) || formalType == typeof(Expression[])) { // Repeated parameter.
                // Make the remaining expressions an array of expressions, possibly empty. Only object[] and Expression[] are supported.
                List<Expression> subList = new List<Expression>();
                for (int j = i; j < actualLength; j++) {
                    subList.Add(actualParamList[j]);
                }
                List<Expression> newParamList = new List<Expression>(); // Build a new list.
                for (int j = 0; j < i; j++) {
                    newParamList.Add(actualParamList[j]); // Copy the first part of the list.
                }
                newParamList.Add(new ParamArrayExpr(subList, tokenFunction));
                actualParamList = newParamList;
                actualLength = formalParams.Length; // Avoid error message that thee ae too many parameters.
                break;
            }
            if (i >= actualLength) { // The formal parameter has no actual parameter.
                error("Only " + actualLength + " parameters. Expects " + formalParams.Length + " parameters.", la); throw new LoggedException();
            }
        }
        if (actualLength > formalParams.Length) { // The formal parameter has no actual parameter.
            error("Too many parameters. Expects " + formalParams.Length + " parameters.", la); throw new LoggedException();
        }
        parameters = actualParamList;
        pi.leftNode = new FunctionCallExpr(pi.leftNode, targetMethod, actualParamList, SQLsupport, tokenFunction);
    } // End functionCall

    void sourceBundle(MethodInfo targetMethod) { // e.g. Me! Bundle(sourceTpl ! findProp)
        Token tokenFunction = token;
        pi.state = PathState.BundleSource; 
        Expect("(");
        Template source;
        if (!pi.Tpl.FormTpl.Templates.TryGetValue(La.val, out source) || source.Ignore) {
            error(la.val + " is not a template in the Form.", la, true); throw new LoggedException();
        }
        pi.Tpl = source; pi.RowTpl = source.RowTpl; // The source template is the result to use in the rest of the path.
        Get(); // Get tpl name.
        if (!IsDot(la, "!", ".")) { error("Expects . or !.", la, true); throw new LoggedException(); }
        Get(); // Get the dot.

        TplPropDesigner designProp;
        ValueInfo propInfo;
        PropertyInfo guiPropInfo;
        Expression exp;
        if (source.DesignProperties.TryGetValue(La.val, out designProp)) {
            exp = new DesignPropRef(pi.leftNode, designProp.cmpPropIndex, designProp, true, token);
        } else
            if (source.Info.ValuePropertyByName.TryGetValue(la.val, out propInfo)) {
                guiPropInfo = (propInfo as ValueInfo).PropertySetter;
                exp = new ValPropertyRef(pi.leftNode, propInfo.propIndex, guiPropInfo, source.valPropArray[propInfo.propIndex], true, token);
            } else { error(la.val + " is not a property in " + source.Name, la, true); throw new LoggedException(); }
        // Cannot check whether the property is a find property, because it may not have been compiled yet. Checked at run time.
        Get(); Expect(")");
        List<Expression> actualParamList = new List<Expression>(1);
        actualParamList.Add(exp);
        pi.leftNode = new FunctionCallExpr(pi.leftNode, targetMethod, actualParamList, false, tokenFunction);

    }


    void defaultExpr(ref Expression e) { // defaultExpr = conditionalExpr { Default conditionalExpr }
        Expression e2 = null;
        conditionalExpr(out e);
        while (La.val.ToUpper() == "DEFAULT") {
            if (pi.inTailExpr) { error("Default is not accepted by the database.", la, true); throw new LoggedException(); }
            Get();
            Token tokenDefault = token;
            pi.state = PathState.Not; 
            conditionalExpr(out e2);
            e = new DefaultExpr(e, e2, tokenDefault);
            if (pi.inTailExpr)
                error("The operator is not allowed in the database. Try: IsNull(a)", tokenDefault);
        }
    }

    void conditionalExpr(out Expression e) { // conditionalExpr =	simpleExpr { ? simpleExpr : conditionalExpr } 
        Expression expTrue = null, expFalse = null;
        pi.state = PathState.Not;
        simpleExpr(out e);
        while (La.val == "?") {
            if (pi.inTailExpr) { error("? : is not accepted by the database. Use iif(a, b, c) instead.", la, true); throw new LoggedException(); } 
            Get();
            pi.brackets.Push(":"); // Stack expected closing token.
            Token tokenQuestion = token;
            pi.state = PathState.Not;
            simpleExpr(out expTrue);
            Expect(":");
            pi.brackets.Pop();
            tokenQuestion = token;
            pi.state = PathState.Not; 
            conditionalExpr(out expFalse);
            e = new CondExpr(e, expTrue, expFalse, tokenQuestion);
        }
    }

    void simpleExpr(out Expression e) { // simpleExpr =	or { Xor or }
        Expression e2 = null;
        Or(out e);
        while (La.val.ToUpper() == "XOR") {
            Get();
            Token tokenXOR = token;
            pi.state = PathState.Not; 
            Or(out e2);
            e = new BinOp(Operator.Xor, e, e2, tokenXOR);
            //if (pathEnded()) pi.state = PathState.Value;
        }
    }

    void Or(out Expression e) { // or =	and { Or and }
        Expression e2 = null;
        And(out e);
        while (La.val.ToUpper() == "OR") {
            Get();
            Token tokenOR = token;
            pi.state = PathState.Not; 
            And(out e2);
            e = new BinOp(Operator.Or, e, e2, tokenOR);
            //if (pathEnded()) pi.state = PathState.Value;
        }
    }

    void And(out Expression e) { // and =	not { And not }
        Expression e2 = null;
        Not(out e);
        while (La.val.ToUpper() == "AND") {
            Get();
            Token tokenAND = token;
            pi.state = PathState.Not; 
            Not(out e2);
            e = new BinOp(Operator.And, e, e2, tokenAND);
            //if (pathEnded()) pi.state = PathState.Value;
        }
    }

    void Not(out Expression e) { // not = [ Not ] compare
        bool bNot = false;
        Token tokenNOT = null;
        if (La.val.ToUpper() == "NOT") {
            Get();
            tokenNOT = token;
            bNot = true;
            pi.state = PathState.Unary; // Intellisense allows Not until a Not has been read.
        }
        Compare(out e);
        if (bNot) {
            e = new UnOp(Operator.Not, e, tokenNOT);
        }
    }

    void Compare(out Expression e) { // compare =	add { [ = | <> | < | > | <= | >= | Like ] add }
        Expression e2 = null; Operator op = Operator.Unknown;
        Add(out e);
        while (true) {
            Token tokenOperator = La;
            if (la.kind == TkKind.Operator) {
                switch (la.val.ToUpper()) {
                    case "=": {
                            Get(); op = Operator.Eq; break;
                        }
                    case "<>": {
                            Get(); op = Operator.Neq; break;
                        }
                    case "<": {
                            Get(); op = Operator.Lt; break;
                        }
                    case ">": {
                            Get(); op = Operator.Gt; break;
                        }
                    case "<=": {
                            Get(); op = Operator.Le; break;
                        }
                    case ">=": {
                            Get(); op = Operator.Ge; break;
                        }
                    default: return;
                } // End switch 
            } // End operator
            else if (la.val.ToUpper() == "LIKE") {
                Get(); op = Operator.Like;
            } else return;
            pi.state = PathState.Not; 
            Add(out e2);
            e = new BinOp(op, e, e2, tokenOperator);
            if (pi.inTailExpr && (e is Constant<object> || e2 is Constant<object>)) 
                error("Null works differently in the database. Comparing with Null will always give null. Try: IsNull(a).", tokenOperator, true);
            //if (pathEnded()) pi.state = PathState.Value;
        }
    }

    void Add(out Expression e) { // add =	mult { [ + | - | & ] mult }
        Expression e2 = null; Operator op = Operator.Unknown;
        Mult(out e);
        while (true) {
            Token tokenOperator = La;
            if (la.val == "-") {
                Get(); op = Operator.Sub;
            } else if (la.val == "+") {
                Get(); op = Operator.Add;
            } else if (la.val == "&") {
                Get(); op = Operator.StrConcat;
            } else return;
            pi.state = PathState.Not; 
            Mult(out e2);
            e = new BinOp(op, e, e2, tokenOperator);
            //if (pathEnded()) pi.state = PathState.Value;
        }
    }

    void Mult(out Expression e) { // mult =	unary { [ * | / | \ | Mod ] unary }
        Expression e2 = null; Operator op = Operator.Unknown;
        Token opToken;
        Unary(out e);
        while (true) {
            if (La.val == "/") {
                Get(); op = Operator.Div; // Always a float result in VB
            } else if (la.val == "*") {
                Get(); op = Operator.Mul;
            } else if (la.val == "\\") { // Truncated divide in VB
                Get(); op = Operator.IntDiv; 
            } else if (la.val.ToUpper() == "MOD") {
                Get(); op = Operator.Mod;
            } else 
                return;
            opToken = token;
            pi.state = PathState.Not; 
            Unary(out e2);
            e = new BinOp(op, e, e2, opToken);
            //if (pathEnded()) pi.state = PathState.Value;
        }
    }

    void Unary(out Expression e) { // unary =	[ + | - ] exponent
        bool bNeg = false, bPos = false;
        Token tokenOperator = La;
        if (la.val == "-") {
            Get(); bNeg = true;
            pi.state = PathState.ExpectPrimary; 
        } else if (la.val == "+") {
            Get(); bPos = true;
            pi.state = PathState.ExpectPrimary; 
        }
        Exponent(out e);
        if (bNeg || bPos) {
            e = new UnOp(Operator.Neg, e, tokenOperator);
        }
    }

    void Exponent(out Expression e) { // exponent =	primary { ^ unary } . Right associative.
        Expression e2 = null;
        Primary(out e); // Allows -2^-3 = -(2^(-3))
        while (La.val == "^") { 
            Get();
            pi.state = PathState.Not; 
            Token tokenOperator = token;
            Unary(out e2); // Right-associative
            e = new BinOp(Operator.Exp, e, e2, tokenOperator);
        }
    }

    void Primary(out Expression e) { // primary =	Constant | ( expression ) | @ primary | path [ '(' SqlExpr ')' ] .
        // Path gets the function parameters too when the path specifies a function property or a system function. 
        // The Sql expression applies when the path specifies an aggregate function. 
        e = null;
        pi.ident = pi.dot = null; // Prevent that Intellisense suggests paths with a left-over initial state.
        if (constant(out e))
            return;
        if (La.val == "(") {
            pi.brackets.Push(")"); // Stack expected closing character.
            Get();
            defaultExpr(ref e);
            Expect(")");
            pi.brackets.Pop(); // Unstack only if there is a matching parenthesis
            return;
        }
        if (La.val == "@") { // @ Calculates a string. In an SQL context, this string is inserted into the SQL string.
            bool oldIsSqlTail = pi.inTailExpr; // Keeps track of whether the path is control-stuff inside an sql tail.
            if (pi.inTailExpr) {
                Expression t = new Constant<string>(MakeSnippet(startSnippet, la) + la.space); // Deal with characters before the @
                if (tail == null) tail = t; else tail = new SQLConcatExpr(tail, t);
                pi.inTailExpr = false;
            }
            Get();
            Expression c; Primary(out c);
            e = new SQL_inject(c);
            pi.inTailExpr = oldIsSqlTail;
            if (pi.inTailExpr) {
                tail = new SQLConcatExpr(tail, e);
                startSnippet = La;
            }
            return;
        }
        if (la.kind == TkKind.Ident || checkDotOps()) { 
            Path(out e); // Gets function parameters and index if they occur in the path (nested paths).
            if (la.val == "(") {
                // Is an aggregate function or a parenthesis after a path error. Check the parameter
                pi.brackets.Push(")"); // Stack expected closing character.
                Get();
                pi.state = PathState.ExpectPrimary; 
                Expression aggrParam = null;
                if (pi.inTailExpr)
                    defaultExpr(ref aggrParam); // We are already in query tail. Handle as other expressions.
                else
                    queryTailExpr(ref aggrParam); // Changes pi as in queryTail. 
                Expect(")");
                pi.brackets.Pop(); // Unstack only if there is a matching parenthesis
                pi.state = PathState.Value; 
                if (pi.aggregateName != null && !pi.inTailExpr) { // If a formula error or inside tail expression, don't create an aggregate field
                    // Should try Aggr1, Aggr2 etc. in case a designer property is called Aggr1 ?? 
                    AggregateField fieldExpr = new AggregateField(pi.aggregateName,
                        "Aggr" + (pi.RowTpl.RowsExpr.aggregateFields.Count + 1).ToString(), aggrParam, la);
                    pi.aggregateName = null; // No more inside aggregate function.
                    pi.RowTpl.RowsExpr.AddAggregateField(fieldExpr);
                }
            }
        } else {
            error(ExpectText(), la); throw new LoggedException();
        }
    } // End primary

    bool constant(out Expression e) { // Constant = long | double | string | time | True | False | Null
        e = null;
        if (La.kind == TkKind.Long) {
            e = new Constant<Int32>((Int32)la.obj, la);
        } else
        if (la.kind == TkKind.Double) {
            e = new Constant<double>((double)la.obj, la);
        } else
        if (la.kind == TkKind.String) {
            e = new Constant<string>((string)la.obj, la);
        } else
        if (la.kind == TkKind.Time) {
            e = new Constant<DateTime>((DateTime)la.obj, la);
        } else
        if (la.val.ToUpper() == "TRUE") {
            e = new Constant<bool>(true, la);
        } else
        if (la.val.ToUpper() == "FALSE") {
            e = new Constant<bool>(false, la);
        } else
        if (la.val.ToUpper() == "NULL") {
            e = new Constant<object>(null, la);
        } else
            return false;
        pi.state = PathState.Value; 
        if (La.error == null || la.cutToken) Get();  // Leave unterminated constant for Intellisense
        else {
            error(la.error, la); throw new LoggedException();
        }

        return true;
    }

    bool checkDotOps() {
        switch (La.val) {
            case ".": return true; // dot
            case "!": return true; // bang
            case "=<": return true;
            case "-<": return true;
            case ">-": return true;
            case ">=": if (pi.state == PathState.RecordSet) return true; else return false;
                // In all other cases, >= is a comparison operator.
            default: return false;
        }
    }

    bool dotOps() { // dotOps = . | ! | =< | -< | >-   Include >= too in state Recordset.
    // When dot-operator: Gets, sets pi.dot and returns true. Else, sets pi.dot = null, and returns false.
        if (la.kind != TkKind.Operator) return false; // 
        if (checkDotOps()) { 
            Get(); pi.dot=token; 
            return true;
        }
        pi.dot = null; // No dot operator after the last ident
        return false;
    }

    //------------------------------------ Query stuff ---------------------------------------------------
    void queryTail(ref Expression e) { // queryTail =	 { Where SqlExpr  |  Group By SqlExpr { , SqlExpr }  |  Having SqlExpr | 
        //  Order By SqlExpr [ Asc | Desc ] { , SqlExpr [ Asc | Desc ] }  |  Top expr | Create expr }  Note: Each clause may occur only once.

        RecordSetExpr recordExpr = e as RecordSetExpr;
        if (recordExpr == null) return; // No record source, no query tail.
        //RecordSetExpr old = tpl.RowsExpr; // Stack old Rows. Important for Intellisense
        tpl.RowsExpr = recordExpr;
        //tpl.RowsCompiled = true; // Consider Rows compiled because the join is okay.
        pi.RowTpl = tpl; // Inside the query tail, the name space includes the Rows.
        try {
            pi.allowSQLkeyword = true;
            while ((La.keyType & KeyType.Sql) > 0) {
                Get();
                //pi.allowSQLkeyword = false;
                switch (token.val.ToUpper()) {
                case "CREATE": if (recordExpr.exprCreate != null) { error("Create occurs twice", token); throw new LoggedException(); }
                    pi.state = PathState.Not; 
                    defaultExpr(ref recordExpr.exprCreate); // Doesn't allow ? : , nor Default              
                    break;
                case "TOP": if (recordExpr.exprTop != null) { error("Top occurs twice", token); throw new LoggedException(); }
                    pi.state = PathState.Not; 
                    defaultExpr(ref recordExpr.exprTop); // Doesn't allow ? : , nor Default              
                    break;
                case "WHERE": if (recordExpr.exprWhere != null) { error("Where occurs twice", token); throw new LoggedException(); }
                    if (recordExpr.exprGroupby.Count > 0) { error("Where must be before Group By", token); throw new LoggedException(); }
                    pi.state = PathState.Not; 
                    queryTailExpr(ref recordExpr.exprWhere);
                    if (formulaError) {
                        recordExpr.exprTop = new Constant<Int32>(1); // Set Top=1 to avoid lots of controls  
                        error("Showing one row as default, since the Where clause has an error", la);
                    }
                    break;
                case "GROUP": if (recordExpr.exprGroupby.Count > 0) { error("Group By occurs twice", token); throw new LoggedException(); }
                    pi.state = PathState.ExpectSymbol; 
                    Expect("By");
                    List<Expression> tempList = new List<Expression> (10); // The group by expression list is built up here so that the expressions don't see the list until it is finished.
                    Expression egroupby = null;
                    pi.state = PathState.Not; 
                    // Group By cannot have @ because it will influence addressing of the row fields.
                    FieldExpr fieldExpr = queryTailExpr(ref egroupby) as FieldExpr;
                    tempList.Add(fieldExpr==null ? egroupby : fieldExpr);
                    while (La.val == ",") {
                        Get();
                        pi.state = PathState.Not;
                        fieldExpr = queryTailExpr(ref egroupby) as FieldExpr;
                        tempList.Add(fieldExpr == null ? egroupby : fieldExpr);
                    }
                    recordExpr.exprGroupby = tempList;
                    //recordExpr.primaryTable = recordExpr.lastTable; // Work around to identify primary tables for Group By.
                    if (recordExpr.exprHaving != null) { error("Group By cannot occur after Having", token); throw new LoggedException(); }
                    if (recordExpr.exprOrderby != null) { error("Group By cannot occur after Order By", token); throw new LoggedException(); }
                    break;
                case "HAVING": if (recordExpr.exprHaving != null) { error("Having occurs twice", token); throw new LoggedException(); }
                    pi.state = PathState.Not; 
                    queryTailExpr(ref recordExpr.exprHaving);
                    if (formulaError) {
                        recordExpr.exprTop = new Constant<Int32>(1); // Set Top=1 to avoid lots of controls
                        error("Showing one row as default, since the Having clause has an error", token);
                    }
                    break;
                case "ORDER": if (recordExpr.exprOrderby != null) { error("Order By occurs twice", token); throw new LoggedException(); }
                    pi.state = PathState.ExpectSymbol; 
                    Expect("By");
                    Expression eorderby = null;
                    pi.state = PathState.Not;
                    queryTailExpr(ref eorderby);
                    PathState old = pi.state; 
                    pi.state = PathState.AscDesc; // To support Intellisense.
                    if ((La.keyType & KeyType.AscDesc) > 0) {
                        Get();
                        eorderby = new SQLConcatExpr(eorderby, new Constant<string>(" " + token.val, token));
                    }
                    pi.state = old;
                    while (La.val == ",") {
                        Get();
                        pi.state = PathState.Not; 
                        eorderby = new SQLConcatExpr(eorderby, new Constant<string>(","));
                        Expression eorderby2 = null; queryTailExpr(ref eorderby2);
                        eorderby = new SQLConcatExpr(eorderby, eorderby2);
                        if ((la.keyType & KeyType.AscDesc) > 0) {
                            Get();
                            eorderby = new SQLConcatExpr(eorderby, new Constant<string>(" " + token.val, token));
                        }
                    }
                    if (!formulaError) recordExpr.exprOrderby = eorderby;
                    break;
                default: error("Unexpected keyword", la); throw new LoggedException();
                }
                if (formulaError) throw new LoggedException(); // Return with the latest pi.state to allow Intellisense.
                pi.allowSQLkeyword = true;
            }
        }
        catch (LoggedException) {
            formulaError = true; tail = null; pi.inTailExpr = true; // The rows must look so that Intellisense can look up field names in the query.
        }
    }


    Token startSnippet, startPath; // The first token in the snippet and the first token of the path. Used inside Path too.
    Expression tail; // The tail element as a concatenation of sql snippets and control property expressions. Used inside Path too.

    Expression queryTailExpr(ref Expression tailElement) { // An sql part such as the A or B-C in Order By A, B-C .
        bool oldInTailExpr = pi.inTailExpr; // queryTailExpr also parses Aggregate parameters - also if the aggregate is in query tail.
        startSnippet = La; pi.inTailExpr = true;
        this.tail = null;
        Expression e = null; // The tail element as an ordinary expression. Not used here. 
        // Generates a classical expression tree in e and a concatenation of sql snippets and control property expressions in tail.
        defaultExpr(ref e); // Doesn't allow ? : , nor Default because inTailExpr is true. Will be reported as an error.
        Expression t = new Constant<string>(MakeSnippet(startSnippet, La)); // Attach the rest of the expression tokens as a snippet.
        if (this.tail == null) this.tail = t; else this.tail = new SQLConcatExpr(this.tail, t);
        tailElement = this.tail;
        if (formulaError) tailElement = null;
        pi.inTailExpr = oldInTailExpr;
        return e;
    }

    string MakeSnippet(Token start, Token top) {
        Token t = start;
        StringBuilder b = new StringBuilder(); //  Add space after each token except at comma/dot.
        while (t != top) {
            b.Append(t.space);
            if ((t.sqlBracket & Token.SqlBracket.left) > 0) b.Append("[");
            b.Append(t.val);
            if ((t.sqlBracket & Token.SqlBracket.right) > 0) b.Append("] ");
            t = t.next;
        }
        return b.ToString();
    }

    void HandleFieldAssignment(Expression e, Token t) { // Adds key fields to the rows query when assigning to a field.
        FieldExpr f = e as FieldExpr;
        if (f == null) return; // Other expressions handle the assignment without row references.
        // Field expressions need special error messages.
        // Because the target has component refs in each row. They all appear "changed" to uVis and all rows would be written to the database. SL 18-05-14.
        if (pi.foreignForm != null) { error("Cannot assign to a field in another form", t); throw new LoggedException(); }
        if (f.rowTpl.RowTpl.RowsExpr.AddKeyFields(f.join) == 0) {
            error("Cannot address the field because it is in a table without a primary key", t); throw new LoggedException();
        }
    }

} // End Class Parser

} // end namespace
