﻿using System;
using System.IO;
using System.Linq;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Data;
using System.ComponentModel;
using System.Reflection;
using System.Data.OleDb;
using VisTool;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Drawing2D;

namespace VisTool {
    public abstract class ContainerComponent : VisComponent {
        // Base class for all canvas-like components such as Form and Canvas. The container holds components such as Textbox (Gui-compoent) and Glyph (shape component). 
        // It receives events for the shapes and routes them to shape.
        // This class is not supposed to be instantiated on its own. Its subclass must add a scrollable guiControl such as Panel or Form.

        public VisComponent[] visComponents = new VisComponent[0];
        public Dictionary<VisComponent, VisComponent> dictComponents = new Dictionary<VisComponent, VisComponent>(100);

        public VisComponent mouseFocus = null; // The component that latest saw the mouse. Used to generate MouseEnter, MouseLeave, MouseHover.
        // The next two variables are for passing events through the z-order stack rather than through canvases (see GetTarget):
        //public int mouseIndex = 0; // The index in visComponents that currently has mouse focus. If the event handler doesn't handle the event, it will be pssed on in z-order sequence.
        //public MouseEventArgs mousePos = null;
        public VisComponent keyFocus = null; // The component that will receive KeyDown, KeyPress and KeyUp.

        public int virtualHeight, virtualWidth; // The size of the virtual canvas. Set during assignCanvas.

        public ScrollableControl scrollControl; // Typed version of the GuiInstance. Set by the subclasses, e.g. VisForm and Canvas. 
        public System.Windows.Forms.Label corner; // Shown in the lower right corner to make the container autoscroll.
        // The position is adjusted at Paint time. This doesn't cause Windows to repeat the paint. However, Studio may cause repaint.

        protected ContainerComponent(int targetCount, DataRow row, Template tpl)
            : base(targetCount, row, tpl) {
            visComponents = new VisComponent [0];
            dictComponents = new Dictionary<VisComponent, VisComponent>();
            corner = new System.Windows.Forms.Label(); // Cannot be inserted in Controls until the Gui ScrollableControl has been created.
            corner.Top = 0; corner.Left = 0;
            corner.Width = 0; corner.Height = 0; corner.Text = "d";
        }

        public override void ResetUpdated() { // Called by kernel at refresh to set dependent properties to not-updated.
            scrollAndOrderUpdated = false;
            virtualHeight = virtualWidth = 0;
            base.ResetUpdated(); 
        }

        public static int AutoScrollID; // Property index. Set at initialization in ReflectDLL. 
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ShowInitially = false,
            ShortHelp = "True when the canvas can scroll to show components outside the visual part. Initially true.", Position = 1)]
        public bool AutoScroll {
            get {
                CmpProp autoScrollProp = this.propertyValues[AutoScrollID];
                if (autoScrollProp.state >= State.Updated) return scrollControl.AutoScroll;
                AutoScroll = toBool(EvalProp(AutoScrollID, true), true);
                return scrollControl.AutoScroll; 
            }
            set { scrollControl.AutoScroll = value; SetUpdated(ref propertyValues[AutoScrollID]); }
        }

        public static int PaddingBottomID; // Property index. Set at initialization in ReflectDLL. 
        public int paddingBottom = 10;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ReadOnly = false, ShowInitially = false, 
            ShortHelp = "The number of pixels from my lowest component to the bottom of the virtual canvas. Initially 10.", Position = 2)]
        public int PaddingBottom {
            get {
                CmpProp paddingBottomProp = this.propertyValues[PaddingBottomID];
                if (paddingBottomProp.state >= State.Updated) return paddingBottom;
                PaddingBottom = toInt32(EvalProp(PaddingBottomID, 0), 0);
                return paddingBottom;
            }
            set { paddingBottom = value; SetUpdated(ref propertyValues[PaddingBottomID]); }
        }

        public static int PaddingRightID; // Property index. Set at initialization in ReflectDLL. 
        public int paddingRight = 10;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ReadOnly = false, ShowInitially = false,
            ShortHelp = "The number of pixels from my rightmost component to the right side of the virtual canvas. Initially 10.", Position = 3)]
        public int PaddingRight {
            get {
                CmpProp paddingRightProp = this.propertyValues[PaddingRightID];
                if (paddingRightProp.state >= State.Updated) return paddingRight;
                PaddingRight = toInt32(EvalProp(PaddingRightID, 0), 0);
                return paddingRight;
            }
            set { paddingRight = value; SetUpdated(ref propertyValues[PaddingRightID]); }
        }

        public static int VScrollPosID; // Property index. Set at initialization in ReflectDLL. 
        public int vScrollPos;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ReadOnly = false, ShowInitially = false, 
            ShortHelp = "The Y position that is at the top of my visible area. Get/Set.", Position = 4)]
        public int VScrollPos {
            get {
                CmpProp vScrollPosProp = this.propertyValues[VScrollPosID];
                if (vScrollPosProp.state >= State.Updated) return vScrollPos;
                VScrollPos = toInt32(EvalProp(VScrollPosID, 0), 0);
                return vScrollPos;
            }
            set { vScrollPos = value; SetUpdated(ref propertyValues[VScrollPosID]);}
        }

        public static int HScrollPosID; // Property index. Set at initialization in ReflectDLL. 
        public int hScrollPos;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ReadOnly = false, ShowInitially = false, 
            ShortHelp = "The X position that is at the left of my visible area. Get/Set.", Position = 5)]
        public int HScrollPos {
            get {
                CmpProp hScrollPosProp = this.propertyValues[HScrollPosID];
                if (hScrollPosProp.state >= State.Updated) return hScrollPos;
                HScrollPos = toInt32(EvalProp(HScrollPosID, 0), 0);
                return hScrollPos;
            }
            set { hScrollPos = value; SetUpdated(ref propertyValues[HScrollPosID]); }
        }

        public virtual string GetCanvasFormula() { // Returns the formula to be used in a component's canvas property in order to place it in this container.
            // Used for instance by drag-and-drop in the container. Example: A PanelGrid cell has this formula: grid1!Cell(i, j).
            return null; // Also null for Forms.
        }

        public virtual void SortShapesByZOrder() {
            visComponents = new VisComponent[dictComponents.Count];
            int i = 0; foreach (var v in dictComponents) visComponents[i++] = v.Value;
            if (visComponents.Length > 1) {
                SortComponents(0, visComponents.Length - 1);
            }
        }

        // Quick Sort Algorithm.
        // Most of the zOrders are 0. Seems stupid to sort them using n * log n comparisons. SL 27-08-14 ??
        // See http://www.sorting-algorithms.com/quick-sort-3-way
        private void SortComponents(int left, int right) { 
            int i = left, j = right;
            int pivot = visComponents[(left + right) / 2].zOrder;
            while (i <= j) {
                //VisEngine.count2 += 2;
                while (visComponents[i].ZOrder < pivot) i++;  // { i++; VisEngine.count2++; }
                while (visComponents[j].ZOrder > pivot) j--; // { j--; VisEngine.count2++; }
                if (i <= j) { // Was i <=j. Don't swop with yourself. SL 19-08-14.
                    VisComponent tmp = visComponents[i];
                    visComponents[i] = visComponents[j];
                    visComponents[j] = tmp;
                    //VisEngine.count1++;
                 i++; j--;
               }
            }
            if (left < i - 1) // Recursive calls
                SortComponents(left, i-1);
            if (i < right)
                SortComponents(i, right);
        }

        public void AddVisComponent(VisComponent visComponent) { // Inserts the component in the canvas and maybe the Gui container.
            /* Components are added and removed one by one to reduce flicker at Refresh. CreateInstanceTree tries to add all of them, 
             * but duplicates are not added. Alternatively, all could be removed at ResetUpdated and inserted with CreateInstance tree without caring for duplicates. 
             */
            VisComponent v;
            if (!dictComponents.TryGetValue(visComponent, out v)) { // Anonymous components may add themselves several times.
                dictComponents.Add(visComponent, visComponent);
            }
            if (visComponent.GuiInstance != null) {
                visComponent.GuiInstance.Top = visComponent.RelativeY(visComponent.Top); // Compensate for present scrolling.  
                visComponent.GuiInstance.Left = visComponent.RelativeX(visComponent.Left); // Compensate for present scrolling.  
                GuiInstance.Controls.Add(visComponent.GuiInstance); // Idempotent
            }
        }

        public void RemoveVisComponent(VisComponent visComponent) { // Inserts the component in the canvas and maybe the Gui container.
            VisComponent v;
            if (dictComponents.TryGetValue(visComponent, out v)) { // Anonymous components may add themselves several times.
                dictComponents.Remove(visComponent);
                if (visComponent.GuiInstance != null)
                    GuiInstance.Controls.Remove(visComponent.GuiInstance);
            }
        }

        public bool scrollAndOrderUpdated = false;

        public override void Paint(PaintEventArgs e) {
            /* Paint is done when all properties in the Form have been evaluated, also those with internal dependencies, e.g. 
             * Bottom or txtBox ! Width. Thus Paint can address properties directly with top, height, etc.
             * 
             * Paint is called not only at refresh, but also at scroll and when another form has hidden part of the VisForm.
             * In these cases scrollAndOrderUpdated is true. As a result SortShapes and finding virtualHeight can be skipped.
             * Windows handles scroll down in this way: It copies the lower part of the canvas to the top of the canvas. Next it 
             * calls paint with the new bottom part as clip area. 
           */
            try {
                VScrollProperties vs = scrollControl.VerticalScroll;
                HScrollProperties hs = scrollControl.HorizontalScroll;
                //Debug.Print("Paint " + this.tpl.Name + ", vs: " + vs.Value + ", clip.Top: " + e.ClipRectangle.Top + ", clip.Height: " + e.ClipRectangle.Height);
                int vMax, hMax; // THe calculated maximum scroll. Zero if no scroll.
                VisForm frm = this.Form(); //if (frm == null) frm = (VisForm)this;
                if (((FormTemplate)frm.tpl).updating)
                    return; // Trying to paint while refresh is in progress.
                if (!scrollAndOrderUpdated) { // && frm.created) {
                    //VisEngine.count1 = VisEngine.count2 = 0;
                    SortShapesByZOrder();
                    corner.Top = PhyS(virtualHeight + paddingBottom) - vs.Value;
                    //Debug.Print(this.tpl.Name + " Virtual+margin-scroll: " + (virtualHeight + bottomMargin - vScrollPos) + " Top: " + corner.Top + " SCrollPos: " + vScrollPos);
                    corner.Left = PhyS(virtualWidth + paddingRight) - hs.Value;
                    vs.SmallChange = 10; vs.LargeChange = Math.Max(10, scrollControl.ClientSize.Height - PhyS(paddingBottom)); // Provide overlap like bottomMargin.
                    hs.SmallChange = 10; hs.LargeChange = Math.Max(10, scrollControl.ClientSize.Width - PhyS(paddingRight));
                    scrollAndOrderUpdated = true;
                } else {
                    //Debug.Print("\nNo sort");
                }

                //Debug.Print("Paint " + this.tpl.Name + " vScroll: " + vScrollPos + " Value: " + (GuiInstance as ScrollableControl).VerticalScroll.Value + " Maximum: " + (GuiInstance as ScrollableControl).VerticalScroll.Maximum);
                if (scrollControl.AutoScroll && vScrollPos != LogS(vs.Value)) { // (GuiInstance as ScrollableControl).AutoScroll && 
                    vMax = Math.Max(0, virtualHeight + paddingBottom - LogS(scrollControl.ClientSize.Height)); // vs.Maximum ignored when AutoScroll is true.
                    if (vScrollPos < 0)
                        vScrollPos = 0;
                    else if (vScrollPos > vMax)
                        vScrollPos = vMax;
                    //ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, this.tpl, "VScrollPosition",
                    //    "Trying vertical scroll to " + vScrollPos + ". Is outside the range 0 to " + vs.Maximum, null);
                    //int old = vs.Value;
                    vs.Value = PhyS(vScrollPos); 
                    //Debug.Print("Scroll old, vScroll, new: " + old + "  " + vScrollPos + "  " + vs.Value + " clip h, w: " + e.ClipRectangle.Height + "  " + e.ClipRectangle.Width);
                }
                if (scrollControl.AutoScroll && hScrollPos != LogS(hs.Value)) {
                    hMax = Math.Max(0, virtualWidth + paddingRight - LogS(scrollControl.ClientSize.Width)); // hs.Maximum ignored when AutoScroll is true.
                    if (hScrollPos < 0)
                        hScrollPos = 0;
                    else if (hScrollPos > hMax)
                        hScrollPos = hMax;
                    hs.Value = PhyS(hScrollPos);
                }

                for (int i = 0; i < visComponents.Length; i++) {
                    VisComponent cmp = visComponents[i];
                    if (!(cmp is ContainerComponent)) {
                        if (cmp.GuiInstance == null)
                            cmp.Paint(e);
                    }
                }
                //base.Paint(e);
            }
            catch (Exception ex) {
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, this.tpl, "", "Unexpected error during Paint: " + ex.Message, null, 0);
            }
        }

        public virtual void OnScroll(object sender, ScrollEventArgs e) {
            if (e.ScrollOrientation == ScrollOrientation.VerticalScroll) VScrollPos = LogS(scrollControl.VerticalScroll.Value);
            else HScrollPos = LogS(scrollControl.HorizontalScroll.Value); 
            // Tempting to use e.NewValue, but it may be outside the min-max range, for instance after a largeScroll at the end;
        }

        public void OnMouseMove(object sender, MouseEventArgs e) {
            //Debug.Print("MouseX e.X: " + SystemItems.MouseX + " " + e.X);
            if (SystemItems.MouseX == this.LogicalX(e.X) && SystemItems.MouseY == this.LogicalY(e.Y)) 
                return; // Windows generates lots of no-moves, for instance when a mouse button is down. Ignore them.
            // MouseMove, etc. needs e.X as well as SystemItems.MouseX to see how far the mouse has moved.

            updatePressed(); // Record AltPressed, etc.
            VisComponent target = getTargetComponent(e);
            if (target != mouseFocus && ! GuiInstance.Capture) { // A component can set Capture and will then keep mouse focus.
                if (mouseFocus != null) mouseFocus.MouseLeave(sender, e);
                mouseFocus = target;
                //Debug.Print("MouseEnter " + target.tpl.Name);
                target.MouseEnter(sender, e);
            }
            if (mouseFocus != null) mouseFocus.MouseMove(sender, e);
            updateMouseVariables(e); // Might be done also in MouseMove, etc. Just to make sure, do it here too.
        }

        public virtual void OnMouseLeave(object sender, EventArgs e) {
            // When the component is on the border of the container, the Gui-mouse may leave the container without a MouseMove out of the component.
            if (mouseFocus != null) mouseFocus.MouseLeave(sender, e);
            mouseFocus = null;
        }

        public void OnMouseHover(object sender, EventArgs e) {
            if (mouseFocus != null) mouseFocus.MouseHover(sender, e);
        }

        public void OnMouseDown(object sender, MouseEventArgs e) {
            if (mouseFocus != null) mouseFocus.MouseDown(sender, e);
        }

        public void OnMouseUp(object sender, MouseEventArgs e) {
            if (mouseFocus != null) mouseFocus.MouseUp(sender, e);
        }

        public void OnGotFocus(object sender, EventArgs e) {
            VisComponent keyFocus = GetKeyFocus(this);
            if (keyFocus != null && keyFocus.GuiInstance == null) keyFocus.GotFocus(sender, e);
        }

        public void OnLostFocus(object sender, EventArgs e) {
            VisComponent keyFocus = GetKeyFocus(this);
            if (keyFocus != null && keyFocus.GuiInstance == null) keyFocus.LostFocus(sender, e);
        }

        public void OnMouseClick(object sender, MouseEventArgs e) {
            VisComponent keyFocus = GetKeyFocus(this);
            if (keyFocus != mouseFocus) {
                if (keyFocus != null && keyFocus.GuiInstance == null) keyFocus.LostFocus(sender, e);
                if (mouseFocus != null) mouseFocus.GotFocus(sender, e);
                keyFocus = mouseFocus;
            }
            if (mouseFocus is VisForm) {
                Form form = mouseFocus.GuiInstance as Form;
                //form.Focus();
                form.Activate(); // Give the form keyboard focus.
            }
            if (mouseFocus != null) mouseFocus.Click(sender, e);
        }

        public void OnMouseDoubleClick(object sender, MouseEventArgs e) {
            VisComponent keyFocus = GetKeyFocus(this);
            keyFocus = mouseFocus;
            if (mouseFocus != null) mouseFocus.DoubleClick(sender, e);
        }

        public void OnKeyDown(object sender, KeyEventArgs e) {
            VisComponent keyFocus = GetKeyFocus(this);
            //Debug.Print("OnKeyDown keyFocus: " + (keyFocus == null ? "null" :
            //    (keyFocus is VisForm ? keyFocus.tpl.Name : keyFocus.Bundle.Form.Name) + "." + keyFocus.tpl.Name));
            if (keyFocus == null) return;
            if (keyFocus.GuiInstance == null || keyFocus is ContainerComponent) {
                // Shapes must get the event. Containers too since they don't handle the events themselves.
                keyFocus.KeyDown(sender, e); // Will update key variables.
                    e.Handled = true;
            }
        }

        public void OnKeyPress(object sender, KeyPressEventArgs e) {
            VisComponent keyFocus = GetKeyFocus(this);
            if (keyFocus == null) return;
            if (keyFocus.GuiInstance == null || keyFocus is ContainerComponent) {
                // Shapes must get the event. Containers too since they don't handle the events themselves.
                keyFocus.KeyPress(sender, e);
                    e.Handled = true;
            }
        }

        public void OnKeyUp(object sender, KeyEventArgs e) {
            VisComponent keyFocus = GetKeyFocus(this);
            if (keyFocus == null) return;
            if (keyFocus.GuiInstance == null || keyFocus is ContainerComponent) {
                // Shapes must get the event. Containers too since they don't handle the events themselves.
                keyFocus.KeyUp(sender, e); // Will update key variables.
                e.Handled = true;
            }
        }

        private VisComponent getTargetComponent(MouseEventArgs e) {
            if (visComponents.Length > 0)
                for (int i = visComponents.Length - 1; i >= 0; i--) { // we start at the tail because it has the highest z order
                    if (visComponents[i].InClickArea(new Point(e.X, e.Y)) && visComponents[i].Visible) {
                        return visComponents[i];
                    }
                }
            return this;
        }

        // The next two methods are for passing events through the z-order stack rather than through canvases:
        //private VisComponent getTargetComponent(MouseEventArgs e) {
        //    for (int i = visComponents.Length - 1; i >= 0; i--) { // we start at the tail because it has the highest z order
        //        if (visComponents[i].InClickArea(new Point(e.X, e.Y)) && visComponents[i].Visible) {
        //            mouseIndex = i; mousePos = e;
        //            return visComponents[i];
        //        }
        //    }
        //    mouseIndex = 0; return this;
        //}

        //public VisComponent getNextTarget() {
        //    for (int i = mouseIndex - 1; i >= 0; i--) { // Continue in the list.
        //        if (visComponents[i].InClickArea(new Point(mousePos.X, mousePos.Y)) && visComponents[i].Visible) {
        //            mouseIndex = i;
        //            return visComponents[i];
        //        }
        //    }
        //    mouseIndex = 0; return null;
        //}

    }

    //------------------------------------------------------------------------------------------------------------------
    public class BufPanel: Panel {
        public BufPanel( ) { 
            DoubleBuffered = VisEngine.doubleBuffer; // For testing new component types.
        } // Must be set in the constructor.
    }

    //------------------------------------------------------------------------------------------------------------------
    [VisClassAtt(ToolboxCategory = "Control", ShortHelp = "Can hold other components and clip them",
        LongHelp = "Can hold other components and clip them", Icon = "canvas.bmp")]
    public class Canvas : ContainerComponent {

        public Canvas(int targetCount, DataRow row, Template tpl)
            : base(targetCount, row, tpl) { // base creates the corner Label.
            Panel panel = new BufPanel();
            this.GuiInstance = panel;
            panel.Tag = this;
            scrollControl = panel;

            panel.AutoScroll = true; // Makes design easier. Disables setting VerticalScroll.Maximum.
            panel.Controls.Add(this.corner);
            this.corner.Parent = panel;
            panel.Cursor = Cursors.Arrow;
            panel.TabStop = false;
            panel.BackColor = System.Drawing.Color.White; // Make the canvas visible. SL 01-06-14
            panel.Paint += OnPaint;

            //AttachControlEvents(panel); // Attach the common control events.
            panel.Scroll += new ScrollEventHandler(OnScroll); // In ContainerComponent.
            panel.MouseEnter += MouseEnter; // In VisComponent
            panel.MouseMove += OnMouseMove; // In ContainerComponent. Finds the VisComponent.
            panel.MouseLeave += OnMouseLeave; // In ContainerComponent
            panel.MouseHover += OnMouseHover;
            panel.MouseDown += OnMouseDown;  // In ContainerComponent. Finds the VisComponent.
            panel.MouseUp += OnMouseUp;
            panel.GotFocus += GotFocus; // Use Activated instead?
            panel.LostFocus += OnLostFocus;
            //panel.MouseClick += FormMouseClick; // Give focus to the Form..
            panel.MouseClick += OnMouseClick; // Select the VisComponent to get uVis focus.
            panel.MouseDoubleClick += OnMouseDoubleClick;
            panel.KeyDown += OnKeyDown;
            panel.KeyPress += OnKeyPress;
            panel.KeyUp += OnKeyUp;
        }

        public void OnPaint(object sender, PaintEventArgs e) {
            e.Graphics.SmoothingMode = SmoothingMode.HighSpeed;
            base.Paint(e);
        }

        public override string GetCanvasFormula() {
            // Returns the formula to be used in a component's canvas property in order to place it in this canvas.
            return this.tpl.Name;
        }

        public static int BorderStyleID; // Property index. Set at initialization in ReflectDLL. 
        private int style;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Special", ReadOnly = false, Position = 2,
            ShortHelp = "The style of my border: 0 (None), 1 (FixedSingle), 2 (Fixed3D). Default: 0 (None)")]
        public int BorderStyle {
            get {
                CmpProp borderStyleProp = this.propertyValues[BorderStyleID];
                if (borderStyleProp.state < State.Updated)
                    BorderStyle = toPosInt32(EvalProp(BorderStyleID, 2), 2);
                return style;
            }
            set {
                (GuiInstance as Panel).BorderStyle =
                (value == 0 ? System.Windows.Forms.BorderStyle.None : value == 1 ? System.Windows.Forms.BorderStyle.FixedSingle : System.Windows.Forms.BorderStyle.Fixed3D);
                style = value; SetUpdated(ref propertyValues[BorderStyleID]);
            }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, null, LongHelp = "Doesn't work for a Canvas. Make it size 0.",
            PropertyCategory = "Common", ReadOnly = false, ShowInitially = false, ShortHelp = "Doesn't work for a Canvas. Make it size 0.", Position = 14)]
        public override bool Visible {
            get { return true; }
            set { }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Common", ReadOnly = false, ShowInitially = false, 
            ShortHelp = "Doesn't work for a Canvas. Will always be black.", Position = 14)]
        public override Color BorderColor {
            get { return Color.Black; }
            set { }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Common", ReadOnly = false, ShowInitially = false, 
            ShortHelp = "Doesn't work for a Canvas. Use BorderStyle.", Position = 14)]
        public override int BorderWeight {
            get { return 1; }
            set { }
        }
    }

    //----------------------------------------------------------------------------------
    public class BufForm : Form {
        public BufForm( ) {
            DoubleBuffered = VisEngine.doubleBuffer; 
        } // Double buffer must be set in the constructor.
    }

    //-------------------------------------------------------------------
    [VisClassAtt(ToolboxCategory = "Control", ShortHelp = "Windows Form. Floats freely on the screen",
        LongHelp = "Windows Form. Floats freely on the screen. Can hold other components and clip them", Icon = "form.bmp")]
    public class VisForm : ContainerComponent {
        public Form form; // The same as GuiInstance, but type-convenient.
        public event EventHandler FormClosed;
        public Point keyFocusPoint; // The point that got the focus. Used for locating deeper controls that might have a key handler.
        //public int keyFocusZindex; // The zOrder index that got the focus. Used for locating deeper controls that might have a key handler.
        Timer timer;

        public VisForm(int targetCount, DataRow row, Template tpl)
            : base(targetCount, row, tpl) {
            form = new BufForm();
            GuiInstance = form;
            GuiInstance.Tag = this;
            Font font = form.Font;
            form.Font = new Font(font.FontFamily, FontSize * SystemItems.Zoom2);
            scrollControl = form;
            top = (int)(GuiInstance.Top/SystemItems.Zoom2); 
            left = (int)(GuiInstance.Left/SystemItems.Zoom2); // Ensure consistency between uVis value and Gui value.
            form.AutoScroll = true; // Makes design easier.
            form.Controls.Add(this.corner);
            this.corner.Parent = form;
            if (tpl.propTop.expr != null || tpl.propLeft.expr != null) // When a position is specified, open the form with "Manual position".
                form.StartPosition = FormStartPosition.Manual; // Do it early or Windows may ignore it.

            form.FormClosed += WindowsFormClosed;
            form.FormClosing += FormClosing; // UVis event handler.
            form.Shown += Shown;
            form.Load += Loaded;
            form.Paint += OnPaint;
            form.Scroll += new ScrollEventHandler(OnScroll); // In ContainerComponent.

            form.KeyPreview = true; // Preview all key events to see whether they go to a Shape.
            form.MouseEnter += MouseEnter; // In VisComponent
            form.MouseMove += OnMouseMove; // In ContainerComponent. Routes it to a VisComponent.
            form.MouseLeave += OnMouseLeave; // In ContainerComponent
            form.MouseHover += OnMouseHover;
            form.MouseDown += OnMouseDown;  // In ContainerComponent. Routes it to a VisComponent.
            form.MouseUp += OnMouseUp; // In ContainerComponent.
            form.GotFocus += OnGotFocus; // In ContainerComponent. Use Activated instead?
            form.LostFocus += OnLostFocus; // In ContainerComponent.
            form.MouseClick += FormMouseClick; // Give focus to the Form ?
            form.MouseClick += OnMouseClick; // then select the VisComponent to get uVis focus.
            form.MouseDoubleClick += OnMouseDoubleClick;
            form.KeyDown += OnKeyDown;
            form.KeyPress += OnKeyPress;
            form.KeyUp += OnKeyUp;
            form.ResizeBegin += OnResizeBegin;
            form.ResizeEnd += OnResized;
            form.Move += Moved;

            timer = new Timer();
            timer.Tick += TimerTick;
        }

        public void OnResizeBegin(object sender, EventArgs e) {
            Form form = this.GuiInstance as Form;
            SystemItems.OldHeight = form.ClientSize.Height;
            SystemItems.OldWidth = form.ClientSize.Width;
            //Debug.Print(this.tpl.Name + "! ResizeBegin old h, w: " + SystemItems.OldHeight + ", " + SystemItems.OldWidth);
        }

        public void OnResized(object sender, EventArgs e) {
            Form form = this.GuiInstance as Form;
            //Debug.Print(this.tpl.Name + "! Resized new h, w: " + form.ClientSize.Height + ", " + form.ClientSize.Width);
            if (SystemItems.OldHeight == form.ClientSize.Height && SystemItems.OldWidth == form.ClientSize.Width)
                return; // Didn't resize. Windows consider any move of the form a resize.
            Resized(sender, e);
        }

        public void FormMouseClick(object sender, MouseEventArgs e) {
            (GuiInstance as Form).Activate();
        }

        public void OnPaint(object sender, PaintEventArgs e) {
            base.Paint(e);
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ShowInitially = false, ReadOnly=true,
            ShortHelp = "The ID number of the form clone. Zero for the first open clone of the form, 1 for the next, etc.")]
        public int CloneID { 
            get { return (this.tpl as FormTemplate).cloneID; }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Common", ShowInitially = true, Position = 5,
            ShortHelp = "The distance from the top of the form to the top of the screen. Default: Window's automatic position.")]
        public override int Top {
            get {
                CmpProp topProp = this.propertyValues[TopID];
                if (topProp.state >= State.Updated) return (int)(GuiInstance.Top/SystemItems.Zoom2);
                object val = EvalProp(TopID, null); // Null when no formula
                if (val != null) Top = toInt32(val, 0);
                return (int)(GuiInstance.Top/SystemItems.Zoom2);
            }
            set { top = value; GuiInstance.Top = (int)(value * SystemItems.Zoom2); SetUpdated(ref propertyValues[TopID]); }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Common", ShowInitially = true,
            ShortHelp = "The height of the form including the bottom border (4 pixels) and title bar (23 pixels).", Position = 6)]
        public override int Height {
            get {
                // Tried to make Height and Width be the client size, but setting the client size = new ... created strange behavior.
                CmpProp heightProp = this.propertyValues[HeightID];
                if (heightProp.state >= State.Updated) return (int)(GuiInstance.Height/SystemItems.Zoom2);
                object val = EvalProp(HeightID, null); // Null when no formula
                if (val != null) Height = toInt32(val, 300);
                return (int)(GuiInstance.Height/SystemItems.Zoom2);
            }
            set { height = value; GuiInstance.Height = (int)(value * SystemItems.Zoom2); SetUpdated(ref propertyValues[HeightID]); }
        }

    [VisPropAtt(VisPropAtt.PropertyType.Property, "' Not supported for Forms.", 
    PropertyCategory = "Common", ReadOnly = true, ShowInitially = false, ShortHelp = "Not supported for Forms.", Position = 7)]
        public override int Bottom { get { return Top + Height; } }

        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Common", ShowInitially = true, Position = 8,
            ShortHelp = "The distance from the left of the form to the left of the screen. Default: Window's automatic position.")]
        public override int Left {
            get {
                CmpProp leftProp = this.propertyValues[LeftID];
                if (leftProp.state >= State.Updated) return (int)(GuiInstance.Left / SystemItems.Zoom2);
                object val = EvalProp(LeftID, null); // Null when no formula
                if (val != null) Left = toInt32(val, 0);
                return (int)(GuiInstance.Left / SystemItems.Zoom2);
            }
            set { left = value; GuiInstance.Left = (int)(value*SystemItems.Zoom2); SetUpdated(ref propertyValues[LeftID]); }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Common", ShowInitially = false,
        ShortHelp = "The width of the the form including the borders (2 * 4 pixels).", Position = 9)]
        public override int Width {
            get {
                // Tried to make Height and Width be the client size, but setting the client size = new ... created strange behavior.
                CmpProp widthProp = this.propertyValues[WidthID];
                if (widthProp.state >= State.Updated) return (int)(GuiInstance.Width/SystemItems.Zoom2);
                object val = EvalProp(WidthID, null); // Null when no formula
                if (val != null) Width = toInt32(val, 500);
                return (int)(GuiInstance.Width/SystemItems.Zoom2);
            }
            set { width = value; GuiInstance.Width = (int)(value * SystemItems.Zoom2); SetUpdated(ref propertyValues[WidthID]); }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, "' Not supported for Forms.",
        PropertyCategory = "Common", ReadOnly = true, ShowInitially = false, ShortHelp = "Not supported for Forms.", Position = 10)]
        public override int Right { get { return Left + Width; } }

        [VisPropAtt(VisPropAtt.PropertyType.Property, "' An open Form is always visible.", Exposed = true,
        PropertyCategory = "Common", ReadOnly = true, ShowInitially = false, ShortHelp = "An open Form is always visible.", Position = 10)]
        public override bool Visible { get { return true; } }

        [VisPropAtt(VisPropAtt.PropertyType.Property, "' A Form cannot be on a Canvas.", PropertyCategory = "Common", ReadOnly = false, ShowInitially = false,
        ShortHelp = "A Form cannot be on a Canvas.", Position = 15)]
        public override ContainerComponent Canvas { get { return null; }
        }

        public override Color backColorDefault { get { return Color.Empty; } }
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Common", ReadOnly = false, ShowInitially = false,
            ShortHelp = "The color of my inner area. Transparent colors not allowed. Default: Light gray, ColorRGB(212, 208, 200).", Position = 11)]
        public override Color BackColor {
            get { return base.BackColor; }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, null, Exposed = false, ShowInitially = false,
            ShortHelp = "The color of the border cannot be set on a VisForm")]
        public override Color BorderColor {
            set { }
        }

        public static int BorderStyleID; // Property index. Set at initialization in ReflectDLL. 
        private int borderStyle;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ReadOnly = false, ShowInitially = true,
            ShortHelp = "The style of the form border: 0 (Fixed size and location), 1 (Fixed size), 2 (Fixed size, 3D border), 3 (Normal, sizable).", Position = 0)]
        public int BorderStyle {
            get {
                CmpProp borderStyleProp = this.propertyValues[BorderStyleID];
                if (borderStyleProp.state >= State.Updated) return borderStyle;
                BorderStyle = toInt32(EvalProp(BorderStyleID, 3), 3);
                return borderStyle;
            }
            set {
                (GuiInstance as Form).FormBorderStyle =
                    (value == 0 ? System.Windows.Forms.FormBorderStyle.None : value == 1 ? System.Windows.Forms.FormBorderStyle.FixedSingle :
                    value == 2 ? System.Windows.Forms.FormBorderStyle.Fixed3D : System.Windows.Forms.FormBorderStyle.Sizable);
                borderStyle = value; SetUpdated(ref propertyValues[BorderStyleID]);
            }
        }

        public static int TextID; // Property index. Set at initialization in ReflectDLL. 
        [VisPropAtt(VisPropAtt.PropertyType.Property, "\"abc\"", ReadOnly = false, ShowInitially = true,
            ShortHelp = "The text shown in the title bar.", Position = 1)]
        public string Text {
            get {
                CmpProp textProp = this.propertyValues[TextID];
                if (textProp.state >= State.Updated) return GuiInstance.Text;
                string textDefault = this.tpl.Name;
                Text = toString(EvalProp(TextID, textDefault), textDefault);
                return GuiInstance.Text;
            }
            set {
                GuiInstance.Text = value; 
                Font font = form.Font;
                form.Font = new Font(font.FontFamily, FontSize * SystemItems.Zoom2);
                SetUpdated(ref propertyValues[TextID]);
            }
        }

        public static int TimerIntervalID; // Property index. Set at initialization in ReflectDLL. 
        public int timerInterval = 0;
        [VisPropAtt(VisPropAtt.PropertyType.Property, "50", ReadOnly = false, 
            ShortHelp = "Number of milliseconds between TimerTick events. When zero, there are no TimerTick events.", Position = 9)]
        public virtual int TimerInterval {
            get { return timerInterval; }
            set {
                if (value <= 0) { if (timerInterval > 0) timer.Stop(); timerInterval = 0; return; }
                if (timerInterval == 0 ) timer.Start(); timer.Interval = timerInterval = value; 
            }
        }

        public static int FontSizeID; // Property index. Set at initialization in ReflectDLL. 
        private float fontSize = 8.50F; // Same as MS default for TextBox and Label.
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, Position = 21, ShowInitially = false,
            ShortHelp = "The font size of my heading cannot be changed. Default: 8.50")]
        public float FontSize {
            get {
                CmpProp fontSizeProp = this.propertyValues[FontSizeID];
                if (fontSizeProp.state >= State.Updated) return fontSize;
                FontSize = (float)toDouble(EvalProp(FontSizeID, 8.50), 8.50);
                return fontSize;
            }
            set {
                fontSize = value;
                Font font = form.Font;
                form.Font = new Font(font.FontFamily, fontSize * SystemItems.Zoom2);
                SetUpdated(ref propertyValues[FontSizeID]);
            }
        }


        [VisPropAtt(VisPropAtt.PropertyType.Eventhandler, null, Position = -10,
        ShortHelp = "Occurs when the Form is loaded. All components have been created, but nothing is shown yet.",
        PropertyCategory = "EventHandler", ReadOnly = false)]
        public void Loaded(object sender, EventArgs e) {
            ExecuteHandler("Loaded", true);
        }

        [VisPropAtt(VisPropAtt.PropertyType.Eventhandler, null, Position = -9,
        ShortHelp = "Occurs when the Form is loaded, shown and got focus.",
        PropertyCategory = "EventHandler", ReadOnly = false)]
        public void Shown(object sender, EventArgs e) {
            ExecuteHandler("Shown", true);
        }

        [VisPropAtt(VisPropAtt.PropertyType.Eventhandler, null, Position = -5,
        ShortHelp = "Occurs when the Form is about to close. If you set System.Handled = true, the Form will not close.",
        PropertyCategory = "EventHandler", ReadOnly = false)]
        public void FormClosing(object sender, FormClosingEventArgs e) {
            this.corner.Focus(); 
            // Change focus so that e.g. text boxes can save their new value. Impossible to set focus to the Form, not even with Form.Activate.
            ExecuteHandler("FormClosing", true);
            e.Cancel = SystemItems.Handled;
        }

        [VisPropAtt(VisPropAtt.PropertyType.Eventhandler, null, Position = -3,
        ShortHelp = "Occurs when the Form's Timer has expired.",
        PropertyCategory = "EventHandler", ReadOnly = false)]
        public void TimerTick(object sender, EventArgs e) {
            ExecuteHandler("TimerTick", true);
        }

        public void WindowsFormClosed(object sender, FormClosedEventArgs e) { // A FormTemplate has only one instance.
            FormTemplate frmTpl = this.tpl.FormTpl;
            foreach (KeyValuePair<string, Template> t in frmTpl.Templates) {
                Template tpl = t.Value;
                if (tpl.RowTpl.RowsExpr != null && tpl.RowTpl.RowsExpr.RecordSet != null) {
                    tpl.RowTpl.RowsExpr.Unload(); // tpl.SQLstring = null; And other flags too.
                    tpl.RowTpl.addedRows.Clear();
                }
            }
            frmTpl.clones[frmTpl.cloneID] = null;
            if (FormClosed != null)
                FormClosed.Invoke(this, new EventArgs()); // E.g. UpdateStudioGrid.

            int openClones = 0;
            foreach (FormTemplate cl in frmTpl.clones) if (cl != null) openClones++;
            if (openClones == 0) { // All clones of this template are now closed. Allow Studio to save the vis-file.
                ErrorLogging.ClearAllErrors(frmTpl.Name);
                frmTpl.clones.Clear();
                //VisEngine.Forms.Remove(frmTpl.Name); Not necessary. Will interupt VisEngine.CloseAllForms.
                int openCount = 0;
                foreach (var cl in VisEngine.Forms) { // Count the open forms excluding those that  are just loaded.
                    List<FormTemplate> clones = cl.Value;
                    for (int i = 0; i < clones.Count; i++)  
                        if (clones[i] != null && clones[i].open) openCount++;
                }
                if (openCount == 0) {
                    // The last form should not close the database connection. The designer may want to open a vis-file.
                    Application.Exit(); // The application, e.g. Simple may also make Exit/ExitThread.
                }
            }
        }
    }

    //---------------------------------------------------
    public class PopupForm : Form {
        VisComponent root;
        const int maxWidth = 500;
        string text;
        const int margin = 3; // Margin inside popup.
        const int padding = 5; //Margin outside root component.
        const int borderWeight = 2;
        RectangleF rect;
        SizeF size;
        StringFormat format = new StringFormat();

        public PopupForm(string text, VisComponent root): base() {
            this.root = root; this.text = text;
            this.ShowInTaskbar = false;
            this.StartPosition = FormStartPosition.Manual; // Do it early or Windows may ignore it.
            this.FormBorderStyle = FormBorderStyle.None;
            this.TopMost = true; // Above normal forms.
            this.BackColor = Color.LightYellow;
            // (root.Bundle.Form.GuiInstance as Form).AddOwnedForm(this); 
            // Makes the popup minimize when the owner is minimized. Doesn't make them move together.
            // However, also makes the root being redrawn at the end of popup, thus hiding dialog windows - annoying.

            Point screenLoc = root.PointToScreen(new Point(0, 0));
            int mouseX = MousePosition.X; int mouseY = MousePosition.Y;

            Graphics dummy = Graphics.FromImage(new Bitmap(1, 1)); // For measuring the text
            size = dummy.MeasureString(text, new Font("Tahoma", 8.5F), maxWidth, format); 
            // Don't scale popup text as e.g. Textbox. Popup isn't part of the graphical layout, but must be very readable.
            size.Height += 2 * (margin + borderWeight);
            size.Width += 2 * (margin + borderWeight);

            if (root.height < root.width) { // Horizontal component. Place popup above component, left of the cursor.
                this.Top = screenLoc.Y - (int)size.Height - padding;
                this.Left = mouseX - (int)size.Width;
            } else { // Vertical component. Place popup to the left of component above cursor.
                this.Top = mouseY - (int)size.Height;
                this.Left = screenLoc.X - (int)size.Width - padding;
            }
            if (this.Top < 0) this.Top = 0; // Move popup inside screen.
            if (this.Left < 0) this.Left = 0;

            this.Height = (int)size.Height; // If we don't set size here, we get a flash in a large area.
            this.Width = (int)size.Width;
            this.Show(); // For strange reasons, we have to set size again. Otherwise small sizes become too large.
            this.Height = (int)size.Height;
            this.Width = (int)size.Width;
            this.rect = new RectangleF(new PointF(margin, margin), size);
        }

        protected override void OnPaint(PaintEventArgs e) {
            // e.Graphics.FillRectangle(new SolidBrush(Color.FromArgb(100, 255, 255, 255)), 0, 0, Width - 2, Height - 2);
            // Attempts to fill with transparent colors makes the entire background default gray.
            e.Graphics.DrawRectangle(new Pen(Color.Black, borderWeight), 0, 0, Width - borderWeight, Height - borderWeight);
            e.Graphics.DrawString(text, new Font("Tahoma", 8.5F * SystemItems.Zoom2), new SolidBrush(Color.Black), rect, format);
            base.OnPaint(e);
        }
    }
}
