﻿// Updated 01:12 29-07-2011 by SL: Template class
using System;
using System.IO;
using System.Linq;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Data;
using System.ComponentModel;
using System.Reflection;
using System.Data.OleDb;
using VisTool;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Drawing2D;


namespace VisTool {

    public enum PropertyKind { GUI, EventHandler, Designer, Standard /* Name, Rows - don't compile as usual */, Function }

    public class TplProperty { // There is a template property for each formula. Some template properties have an empty formula, 
        // a formula that consists of a comment only, or an erroneous formula. These template properties have expr == null.
        public Template tpl; // The template where this template property belongs.
        public string Name; // Makes it easy to get the property name when only the template property ref is available.
        public string Formula = null; // The formula has trimmed end, i.e. no EOL. For event handlers, the formula is a statement list.
        // Studio shows properties that have a non-null formula (e.g. "" or "10") or are ReadOnly.
        public Token startChain = null; // Allows access to all tokens in the scanned formula. Terminates with an endless loop of EOL-tokens.
        public List<TokenError> tokenErrors = new List<TokenError>(0); // List of (error, token) for this property
        public Expression expr = null; // Is null if there isn't a formula or the formula is empty. 
        // Is an ordinary expression if the formula has only one expression. The expression has init = true or false.
        // Is an expression list with at least two elements when the formula is a list, e.g. Timescale borderValues. 
        // Each element in the expression list is null (empty formula) or an expression with init = true or false.
        // Ordinary expressions match simple properties as well as list properties (become a value list with one elelment). 

        //public PropertyInfo PropertyGUISetter; Done through PropInfo.
        public int cmpPropIndex = -1; // Refers to the value in propertyValues[i] or designerPropValues[i]. 
        // Only properties with a formula (maybe null expression) and designer properties have a ControlPropery.
        public Type PropertyValueType; // E.g. Int32, string, Color, Font
        public PropertyKind PropKind; // SL: messy names. Replace with PropInfo.Kind: GUI, Extended, EventHandler, Designer, Standard, Function ...

        public Dictionary<List<object>, Bundle> FindBundles = null; // When the property formula is Find tplA Where . . . , this is a dictionary of all tplA components. Each entry refers to a bundle of tpl components.

        public TplProperty(Template tpl, string name, Type propValueType, PropertyKind propKind) {
            this.tpl = tpl;
            this.Name = name;
            this.PropertyValueType = propValueType;
            this.PropKind = propKind;
        }

        public static TplProperty GetTemplateProperty(Template tpl, string propName, out PropInfo propInfo) {
            // Returns null when there is no such property or a function property. Returns propInfo = null if it is a designer property.
            if (tpl.Info.PropertyByName.TryGetValue(propName, out propInfo)) {
                if (propInfo is ValueInfo)
                    return tpl.valPropArray[propInfo.propIndex];
                if (propInfo is EvtInfo)
                    return tpl.evtPropArray[propInfo.propIndex];
                return null; // A function property
            }
            TplPropDesigner designProp;
            if (tpl.DesignProperties.TryGetValue(propName, out designProp)) {
                propInfo = null; return designProp;
            }
            else return null;
        }
    } // end template property

    public class TplPropValue : TplProperty {
        public ValueInfo PropInfo;
        public TplPropValue(Template tpl, string name, Type propValueType, PropertyKind propType) : base(tpl, name, propValueType, propType) { }

        // public virtual object EvalAndSet(VisComponent ctrl) { // Never called - replaced by EvalProp

        public virtual void SetValue(VisComponent ctrl, object o, CmpProp prop) { // Used only by EvalAndSet. 
            // Might check for value changed if we keep the value in prop.
                this.PropInfo.PropertySetter.SetValue(ctrl, o, null);
        }

        private object[] getListValue(object oldValue, Type type, int count) {
            object[] result = new object[count];
            for (int i = 0; i < count; i++)
                result[i] = type.GetMethod("get_Item").Invoke(oldValue, new object[] { i });
            return result;
        }
    } // end template property

    public class TplPropDesigner : TplProperty { 
        public TplPropDesigner(Template tpl, string name, Type propValueType, PropertyKind propType) : base(tpl, name, propValueType, propType) { }

        public object EvalAndSet(VisComponent ctrl) { // Return value never used ! ??
            CmpProp prop = ctrl.designerPropValues[this.cmpPropIndex];
            if (prop.state >= State.Updated) return prop.Value;
            if (this.expr == null)
                return null;
            if (this.expr is ExpressionList) {
                if (prop.Value == null) prop.Value = new object[0];
                return ctrl.EvalListProp(this, prop, prop.Value); // Checks Visited.
            }

            // There is an expression. Evaluate it.
            if (prop.state == State.Visited) {
                VisEngine.circularRef = prop; VisEngine.circularLevel = 0;
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, tpl, this, "End of circular reference, level 0", null, ctrl.Index());
                return prop.Value;
            }
            prop.state = State.Visited;
            try {
                object o = this.expr.Eval(ctrl, tpl);
                prop.Value = o;
                prop.state = expr.Init ? State.Initialized : State.Updated;
                if (VisEngine.circularRef == prop)
                    VisEngine.circularRef = null; // I am the start and the end. The circle is closed.
            }
            catch (Exception ex) { // All exceptions thrown by expression nodes are logged here.
                prop.state = State.Updated; // Don't propagate errors.
                Token t = VisEngine.curExprToken;
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, tpl, VisEngine.curTplProp, ex.Message, t, ctrl.Index());
            }
            return prop.Value;
        }
        
        public void SetValue(VisComponent ctrl, object o, CmpProp prop) {
            ctrl.designerPropValues[this.cmpPropIndex].Value = o;
        }
    }

    public class TplPropEvent : TplProperty {
        public EvtInfo PropInfo;
        public TplPropEvent(Template tpl, string name, Type propValueType, PropertyKind propKind) : base(tpl, name, propValueType, propKind) { }

        public object EvalAndSet(VisComponent ctrl) { return null; } 
        // An event property has no value and doesn't create event handlers. It is just a place-holder for the statement formula.
    }

    public class CmpProp { // A value property, either built-in property or design property.
        public object Value = null; // For designer properties: The current value of the property. For built-in properties: Intended for optimization, but not used consistently.
        // Could look like this in VisComponent properties: instProp.Value = top = ConvertTo ...;
        // For Find-properties, the value is a BundleRef: (bundle, indexInBundle)
        public TplProperty tplProp; // Allows easy access to property kind, value type, etc.
        public VisComponent cmp; // Allows easy access to the component, e.g. for Find bundles.
        public State state = State.Old; // Old: The value has to be calculated. Visited: Calculation is in progress. Updated: The value is ok, but has to be recalculated
        // at refresh. Initialized: The value was calculated. User actions may change it. Don't recalculate at refresh.

        public CmpProp(TplProperty tplProp, VisComponent cmp) {
            this.tplProp = tplProp; this.cmp = cmp;
        }
    }

    public enum State { Old, Visited, Updated, Initialized }; // Old: The value has to be calculated. Visited: Calculation is in progress. Updated: The value is ok, 
    // but has to be recalculated at refresh. Initialized: The value was calculated. User actions may change it. Don't recalculate it at refresh.
    // A list property is never Initialized, only Updated. It will be recalculated but its Init-elements are not recalculated in soft refresh modes (ValueChange and Requery).

    public class BundleRef { // A Find result is a BundleRef. The bundle contains all source components with the same target component.
        public Bundle bundle; // A bundle with the target component as Target. Sufficient for simple formulas like findProp !Top
        public int index; // The index in the bundle where the source component is. Needed for formulas like findProp[index-1] !Top
        public object[] orderKeys; // The order keys for the source component.
        public object tag; // Some components may add their own information to the BundleRef, e.g. Tree.

        public BundleRef(Bundle bundle, int index, object[] orderKeys) {
            this.bundle = bundle; this.index = index; this.orderKeys = orderKeys;
        }
    }

    public enum AccessMode { Hidden, ReadOnly }

    public struct PropSpec {
        public string PropName;
        public AccessMode Mode;
    }

    public class Template {
        // The Rows property may be compiled out of order because other templates refer to its fields.
        // For an ordinary query, only the join part is needed for this. Since it doesn't contain references to other templates, no cyclical references can occur.
        // However, when it is a Group By, the entire query must be compiled to identify the fields and possible aggregate functions. Cyclical references can thus occur.
        // RowsCompiled and Visited detects this.
        public bool RowsCompiled = false; // Formula has been compiled.
        public bool Visited = false; // Rows being compiled or instantiated.
        public bool Updated = false; // True when the template has been instantiated.
        public bool Ignore = false; // When true, the template can be edited, but it is not compiled or instantiated. Set with Name property "nn Ignore". 

        public Template RowTpl; // The template that holds the rows for this template. Self ref unless the rows formula is a template ref.
        public FormTemplate FormTpl; // The Form template has itself as FormTpl.
        public List<System.Data.DataRow> addedRows = new List<System.Data.DataRow>(); // Keeps track of new rows to be written to the database at commit. Usually only one row.

        public List<TplProperty> findProperties = new List<TplProperty>(4); // List of the find properties (tplName ! propertyName) that refer to this template.
        // The compiler sets up the list. The list matches List findBundles in VisComponent.
        public Dictionary<string, TplPropDesigner> DesignProperties;
        public TplPropValue[] valPropArray; // The value properties structured as an array.
        public TplPropEvent[] evtPropArray; // The event properties structured as an array.
        // Value properties and Designer properties address component properties by means of the index in propArray.
        public ComponentInfo Info; // Contains template Type and dictionaries of properties.

        public string Name; // Also exists as a property:  tpl.valPropArray[VisComponent.NameID].  
        // A NameExpression handles references to the Name "property". SL 15-04-13 
        // In case of errors, the name in the property doesn't match Name in the template.
        public string Comment; // SL: Contains the space text preceding the first property in a vis-file.

        public TplProperty propRows, propTop, propHeight, propBottom, propLeft, propWidth, propRight, propCanvas; // For easy access to these properties,
        // e.g. to see whether they have a formula. The cumbersome way is: propRows = valPropArray[VisComponent.RowsID];

        public string SQLstring = null; // Latest SQL string. Used to check for changes at Refresh
        public int createCount = 0; // Latest createCount. Used to check for changes at Refresh.
        //public bool SQLchanged = true; // Whether the sql has changed during the update. For optimizing linked recordsets.

        public RecordSetExpr RowsExpr = null; // TODO: SL: Null until Rows has been compiled. Null if the template has no Rows.
        public Bundle RowBundle; // The bundle of the template's components. Templates with shared Rows have separate RowBundles.

        public Template(string name, ComponentInfo cmpInfo, FormTemplate frmTpl) {
            this.Name = name;
            this.RowTpl = this;
            this.FormTpl = frmTpl; // Don't set the ParentTemplate until the right one is known. CreateTemplateTree uses it to see whether setting is necessary.
            CreateTplProps(cmpInfo);
            //RowBundle = new Bundle(0, null, this.propRows); // Allows easy loops without testing for Count == 0.
        }

        public void CreateTplProps(ComponentInfo cmpInfo) { // Used also by ChangeTemplateType.
            this.Info = cmpInfo;
            this.valPropArray = new TplPropValue[cmpInfo.ValuePropertyByName.Count];
            foreach (var item in cmpInfo.ValuePropertyByName) {
                ValueInfo valInf = item.Value;
                TplPropValue tplProp = new TplPropValue(this, valInf.Name, valInf.Type, PropertyKind.GUI);
                tplProp.cmpPropIndex = valInf.propIndex;
                tplProp.PropInfo = valInf;
                valPropArray[valInf.propIndex] = tplProp;
            }
            propRows = valPropArray[VisComponent.RowsID]; propCanvas = valPropArray[VisComponent.CanvasID];
            propTop = valPropArray[VisComponent.TopID]; propHeight = valPropArray[VisComponent.HeightID]; propBottom = valPropArray[VisComponent.BottomID];
            propLeft = valPropArray[VisComponent.LeftID]; propWidth = valPropArray[VisComponent.WidthID]; propRight = valPropArray[VisComponent.RightID];
            propRows.PropKind = valPropArray[VisComponent.NameID].PropKind = PropertyKind.Standard;

            this.evtPropArray = new TplPropEvent[cmpInfo.EventByName.Count];
            foreach (var item in cmpInfo.EventByName) {
                EvtInfo evtInf = item.Value;
                TplPropEvent tplProp = new TplPropEvent(this, evtInf.Name, null, PropertyKind.EventHandler);
                tplProp.cmpPropIndex = evtInf.propIndex;
                tplProp.PropInfo = evtInf;
                evtPropArray[evtInf.propIndex] = tplProp;
            }
            DesignProperties = new Dictionary<string, TplPropDesigner>(5, StringComparer.OrdinalIgnoreCase);
        }

        public virtual void deleteInstances() { // Used for Forms too ??
            if (this.RowBundle == null) return; // Test for bundle == null if used for Forms ?
            foreach (CmpProp cmpProp in this.RowBundle.Instances) { 
                cmpProp.cmp.Dispose(); // Disposes GuiInstance too.
            }
            RowBundle.Instances.Clear();
            if (this.RowTpl.RowsExpr != null) {
                this.RowTpl.RowsExpr.Unload(); // Delete loaded tables
                this.RowTpl.addedRows.Clear();
            }
        }

        public virtual void deleteInstances(Bundle bundle, int first, int count) { // Delete a range of instances. Keep the Rows.
            for (int i = first; i < first + count; i++) {
                VisComponent inst = bundle.Instances[i].cmp;
                inst.Row = null; // Delete reference from instance to row.
                inst.Dispose(); // Disposes GuiInstance too.
                bundle.Instances[i] = null;
            }
        }
    }

    public class FormTemplate : Template {
        public int cloneID;
        public List<FormTemplate> clones = new List<FormTemplate>(10); // Each clone has its own parameters and its own bundle of Forms.
        public bool updating = false; // True while this bundle of forms is being refreshed.
        public bool open = false; // False when the form is loaded but not opened. This happens at crossrefs during compilation. 
        public Expression[] exprParameters = null; // Evaluated at FormOpen. May later be used for assignment. 
        public object[] parameters = null; // Holds the evaluated parameters.
        public VisComponent callingContext; // Used for Assignment. Risk of dangling contexts when the caller is closed before this form.

        // Dictionary of all templates in the Form including the form itself and including Ignored templates:
        public Dictionary<string, Template> Templates = new Dictionary<string, Template>(StringComparer.OrdinalIgnoreCase);
        // Templates that have dirty data or added/deleted/modified rows:
        public Dictionary<string, Template> DirtyDataTemplates = new Dictionary<string, Template>(StringComparer.OrdinalIgnoreCase);

        public bool isDirty = false; // True when a form formula has been changed. Set only when just one clone is open, 

        public FormTemplate(string name)
            : base(name, VisEngine.ComponentByName["VisForm"], null) {
            this.Name = name; // Also set in base constructor.
            this.Info = VisEngine.ComponentByName["VisForm"];
            this.Templates.Add(name, this); // Add the form to its own dictionary of templates.
            this.FormTpl = this;
        }
    }

    public class Bundle {
        public int Last { get { return this.Instances.Count - 1; } } // The last valid index. Delete ??
        public List<CmpProp> Instances;
        public State orderState = State.Old; // Old: Has to be filled and/or sorted. Visited: Sorting in progress. Updated: Fill and sorting ok. 
        public VisComponent Target; // For Find bundles, the target component, for Rows null.
        public TplProperty tplProp;

        public Bundle(int numOfInstances, VisComponent Target, TplProperty tplProp) {
            this.Target = Target; this.tplProp = tplProp;
            this.Instances = new List<CmpProp>(numOfInstances);
        }
    }

}
