﻿using System.Linq;
using System.IO;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Drawing;
using System.Data;
using System.ComponentModel;
using System.Reflection;
using System.Globalization;

using System.CodeDom.Compiler;
using Microsoft.CSharp;
using System.Xml.Linq;
using System.Xml.XPath;
using System;
using System.Diagnostics;
using VisTool;

namespace VisTool {

    public partial class VisEngine {
        public static readonly string ProgramDirectory = Directory.GetCurrentDirectory(); // The start-up folder where .exe, .dll, icons, etc. reside.
        public static string ApplicationDirectory = ""; // The folder where .vis, .vism, application icons, map settings, etc. reside.
        public static string MapFile = null; // The most recent map file opened.

        // MODES FOR THE ENGINE loaded from .vism file
        public enum interactionMode { AsEndUser, AsDesigner, DesignWithControl, DesignWithAlt }
        public static interactionMode InteractionMode = interactionMode.AsEndUser;
        public static bool LimitedRows = false; // When true, MaxRows take effect.
        public static int MaxRows; // zero to anything, where zero means always one
        public static bool doubleBuffer = true; // When true, Paint is done in a another screen buffer, then shown at once. 
        // doubleBuffer false is convenient when developing new components and see how they paint step by step.
        public static bool simulatedTime = false;
        public static DateTime simulatedDateTime;
        public static double simulatedTimeIncrement;
        public static bool simulatedUpdate = false;
        public static bool intellisensePopUpInitially = false; // This mode is not visible on Studio panel. SL 09-03-13.
        public static int screenWidth = 0; // where 0 is default
        public static int screenHeight = 0; // where 0 is default
        public static string startUpForm;
        public static float Zoom;
        //Design mode
        public static event EventHandler<DesignModeEventArgs> ShowForm;

        // Dictionaries of all Components and System items: They are initialized in the ComponentInfo constructor during a scan of the .dll files.
        public static Dictionary<string, ComponentInfo> ComponentByName = new Dictionary<string, ComponentInfo>(30, StringComparer.OrdinalIgnoreCase);
        public static Dictionary<string, SystemMethodInfo> SystemMethodByName = new Dictionary<string, SystemMethodInfo>(100, StringComparer.OrdinalIgnoreCase);
        public static Dictionary<string, SystemFieldInfo> SystemFieldByName = new Dictionary<string, SystemFieldInfo>(30, StringComparer.OrdinalIgnoreCase);
        // In case we need signature overload, the following Dictionary may be useful:
        //public static Dictionary<string, List<SystemMethodInfo>> SystemMethodByName = new Dictionary<string, List<SystemMethodInfo>>(StringComparer.OrdinalIgnoreCase); This dictionary could be used in the future instead to support method overloading

        // Dictionaries of all forms in the application directory: 
        public static Dictionary<string, List<FormTemplate>> Forms = new Dictionary<string, List<FormTemplate>>(100, StringComparer.OrdinalIgnoreCase);
        public static List<KeyValuePair<string, List<FormTemplate>>> SortedForms; // For Intellisense.
        public static Dictionary<string, string> FormNames; // The key is a form name, the value the true short file name.

        public static CultureInfo codeCulture = new CultureInfo("en-gb"); // Use European English number formats and dates in formulas, WindowsPositions, etc..
        public static int ErrorCount = 0; // Error handling takes time. Stop calculation, etc. when too many errors.
        public const int MaxErrors = 50; // Stop reporting after this number of distinct errors.
        public static bool MaxErrorReported = false; // True when VisEngine has reported that there are too many errors.

        static VisEngine() { // Static constructor. Initializes the dictionaries.
            // The following is for documentation of scaling. Can be removed.
            ErrorLogging.Log("Scale data: "); // Initiallizes errorlogger.
            Graphics dummy = Graphics.FromImage(new Bitmap(1, 1));
            Debug.Print("Most screens have 96 Dots Per Inch (96 DPI). This screen has DpiX: " + dummy.DpiX + ", DpiY: " + dummy.DpiY);
            StringFormat format = new StringFormat();
            FontFamily fontFamily = new FontFamily("Tahoma");
            Font font = new Font(fontFamily, 10F, FontStyle.Regular);
            string test = "A";
            SizeF size = dummy.MeasureString(test, font, 1000, format);
            Debug.Print("A 10 point A has this pixel size: Width: " + size.Width.ToString(codeCulture) + ", height: " + size.Height.ToString(codeCulture));

            ComponentInfo.ReflectDLLs(); // Initializes component data and system methods.
        }

        public static void UpdateFormDirectories(string oldName, string newName) {
            Dictionary<string, List<FormTemplate>> newForms = new Dictionary<string, List<FormTemplate>>(100, StringComparer.OrdinalIgnoreCase);
            DirectoryInfo dir = new DirectoryInfo(VisEngine.ApplicationDirectory);
            FileInfo[] inf = dir.GetFiles("*.vis");
            for (int f = 0; f < inf.Length; f++) {
                string trueName = inf[f].Name;
                if (inf[f].Extension.ToUpper() != ".VIS")
                    continue; // Extensions like "vism" and "visER" are returned with GetFiles !?
                // Has correct case, whereas new FileInfo("clients"), returns filename "clients" although the true file name is "Clients".
                List<FormTemplate> item = null;
                string frmName = trueName.Remove(trueName.Length - 4);
                if (Forms.TryGetValue((frmName == newName ? oldName : frmName), out item))
                    newForms.Add(frmName, item); // Copy any open form templates to newForms.
                else
                    newForms.Add(frmName, new List<FormTemplate>(5)); 
            }
            VisEngine.Forms = newForms; // Use the new form dir.
            SortedForms = newForms.ToList();
            SortedForms.Sort((a, b) => (a.Key.CompareTo(b.Key)));
            FormNames = new Dictionary<string, string>(Forms.Count, StringComparer.OrdinalIgnoreCase);
            foreach (var item in newForms) FormNames.Add(item.Key, item.Key); // Create dictionary of true file names (correct caps).
        }

        public static FormTemplate GetFormTemplate(string formName, ref int cloneID) {
            // Returns the Form template with this name and cloneID if it is open. 
            // Returns the Form template with the first cloneID that is open.
            // Returns null and cloneID = 0 if the form isn't open.
            // Returns null and cloneID < 0 if the form doesn't exist as a formName.vis file.
            if (Forms.Count == 0) // First form not opened. Record all vis-forms in Forms.
                UpdateFormDirectories(null, null);
            List<FormTemplate> clones = null;
            if (!VisEngine.Forms.TryGetValue(formName, out clones)) { cloneID = -1; return null; } // Form doesn't exist as a file.
            if (cloneID >= 0 && cloneID < clones.Count && clones[cloneID] != null) return clones[cloneID]; // A form with this cloneID exists. Return it.
            for (int i = 0; i < clones.Count; i++)
                if (clones[i] != null && clones[i].open) { cloneID = i; return clones[i]; } // Found another open clone.
            cloneID = 0; return null; // The form exists but isn't opn.
        }

        // List of all open DataMaps. At present the kernel supports only one, but the VismParser can support several.
        public static List<DataMap> DataMaps = new List<DataMap>();

        public static Stopwatch sw1 = new Stopwatch(); // For measurements
        public static Stopwatch sw2 = new Stopwatch();
        public static long count1 = 0, count2 = 0;
        public static long s1 = 0, s2 = 0, s3 = 0, s4 = 0, s5 = 0, s6 = 0; // Elapsed times on the way

        public static VisComponent curComponent; // Holds the component being evaluated. Used to identify function callers.
        public static Token curExprToken; // Holds the token that is the source of the current expression node. Used for error messages during evaluation.

        // curTemplate and curTplProp are set when evauation of a formula starts: In CalculateProperties and in expression nodes that refer to property.
        // Error messages will be connected to this property.
        public static Template curTemplate; // Holds the template being evaluated. Used for error messages.
        public static TplProperty curTplProp; // Holds the tplProp being evaluated. Used for error messages.

        /* CIRCULAR REFERENCES: When evaluation of a property detects that the property is already being evaluated (Visited), it logs it as an error 
        * (End of circular reference), sets circularRefEnd and returns with a default value.
        * An expression that refers to a property, checks for circularRefEnd <> null, logs it as an error and returns the default value. 
        * When the expression evaluates the circularRefEnd, we are at the root af the circle and circularRefEnd is set to null.
         */
        public static CmpProp circularRef = null; // The component property that was Visited when VisEngine tried to evaluate it.
        public static int circularListIndex = 0; // When the property value at the cyclical end is a list, the index in the list that was Visited. 
        public static int circularLevel; // The number of reference steps away from circularRef.
        public static Int32[] loopVariables = new Int32[10]; // Holds the loop variables in nested For loops.

        public enum RefreshMode { ValueChange, Requery, FormulaChange }; // Sequence important for testing. 
        public static RefreshMode refreshMode = RefreshMode.ValueChange;
        public static bool CompileOnly = false; // When true, compiles, but doesn't delete or change components. Used by Studio when editing many formulas.
        /*
         * Initially refreshMode is ValueChange - the most modest refresh. As changes to formulas and templates accumulate, refreshMode
         * increases. Refresh() refreshes parts according to refreshMode. At the end of refresh or OpenForm, refreshMode is reset to ValueChange.
        */

        public static BundleRef leftBundleRef = null; // Index and Last use this in leftPath[Index] to refer to the index in the path rather than to Me ! Index.
        // component.Find( b ) uses leftPath to make b refer to component[index] rather than component[0].

        public static VisComponent SelectedComponent = null;

        public static void CompileRows(FormTemplate frm) { // Compiles all rows formulas. 
            //Debug.Print("CompileRows " + frm.Name);
            // Necessary to rebuild the field list, join list, etc. which may change as the result of other formulas referring to another field.
            foreach (KeyValuePair<string, Template> t in frm.Templates) {
                Template tpl = t.Value;
                if (tpl.RowsCompiled || tpl.Ignore) continue;
                Parser p;
                if (tpl.propRows.Formula == null) continue; // Null rows formula possible. Not accepted by Parser( ).
                p = new Parser(tpl, new StringReader(tpl.propRows.Formula), tpl.propRows, false); // Uses or sets startChain. Sets refs for error handling.
                p.rowsFormula(tpl.propRows); // Compiles the Rows formula and sets tpl.ParentTemplate. 
            }
        }

        public static void CompileTemplates(FormTemplate frmTpl) {
            //Debug.Print("CompileTemplates " + frmTpl.Name);
            // For each template and each template property, compile the formula and set TemplateProperty.expr.
            Parser parser;

            foreach (var item in frmTpl.Templates) {
                Template tpl = item.Value;
                if (tpl.Ignore) continue;
                foreach (TplProperty tplVal in tpl.valPropArray) { // Compile value properties.
                    if (tplVal.cmpPropIndex == 0) continue; // Name property
                    if (tplVal == tpl.propRows) continue; // Don't compile Rows. Handled by CompileRows or PrepareRowTpl. 
                    if (tplVal.Formula == null) continue;
                    parser = new Parser(tpl, new StringReader(tplVal.Formula), tplVal, false);
                    parser.Formula(tplVal); // Parser catches errors
                }
                foreach (TplProperty tplEvt in tpl.evtPropArray) { // Compile event properties.
                    if (tplEvt.Formula == null) continue;
                    parser = new Parser(tpl, new StringReader(tplEvt.Formula), tplEvt, false);
                    parser.Formula(tplEvt); // Parser catches errors
                }
                TplPropDesigner tplProp;
                foreach (var e in tpl.DesignProperties) { // Compile designer properties.
                    tplProp = e.Value;
                    if (tplProp.Formula == null) continue;
                    parser = new Parser(tpl, new StringReader(tplProp.Formula), tplProp, false);
                    parser.Formula(tplProp); // Parser catches errors
                }

                if (tpl.propTop.expr != null && tpl.propHeight.expr != null && tpl.propBottom.expr != null) {
                    // All three specified. Ignore Top.
                    tpl.propTop.expr = null;
                    ErrorLogging.Log(ErrorMode.CompileTime, Priority.Warning, tpl, tpl.propTop, "Ignoring Top since Bottom and Height are defined.", null, 0);
                }
                if (tpl.propLeft.expr != null && tpl.propWidth.expr != null && tpl.propRight.expr != null) {
                    // All three specified. Ignore Left.
                    tpl.propLeft.expr = null;
                    ErrorLogging.Log(ErrorMode.CompileTime, Priority.Warning, tpl, tpl.propLeft, "Ignoring Left since Width and Right are defined.", null, 0);
                }
            }
        }

        static void CalculateProperties(Template tpl, VisComponent inst, int index) {
            curTemplate = tpl; // Prepare for errors.
            //Dictionary<string, TemplateProperty>.Enumerator e = tpl.Properties.GetEnumerator();
            object obj;
            foreach (TplPropValue tplProp in tpl.valPropArray) {
                //if (tplProp.PropInfo == null) continue; // Name and other Ignore properties have null PropInfo.
                try {
                    curTplProp = tplProp; // Name and Rows return a valye, but don't set anything.
                    obj = tplProp.PropInfo.PropertySetter.GetValue(inst, null);
                    if (circularRef != null) { // The start point of the circular reference. The error has been logged.
                        //ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, tpl, curTplProp, "Start of circular reference, level " + ex.level.ToString(), null);
                        circularRef = null;
                    }
                }
                catch (Exception ex) { // All exceptions thrown by value properties are logged here.
                    Token t = VisEngine.curExprToken;
                    ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, tpl, curTplProp, "Should have been caught by EvalProp: " +
                        (ex.InnerException != null ? ex.InnerException.Message : ex.Message), t, inst.Index());
                    continue;
                }
            }
            //inst.propertyValues[VisComponent.RowsID].Value = new BundleRef(tpl.RowBundle, index, null);

            foreach (var e in tpl.DesignProperties) {
                try {
                    curTplProp = e.Value;
                    (curTplProp as TplPropDesigner).EvalAndSet(inst);
                    if (circularRef != null) { // The start point of the circular reference. The error has been logged.
                        //ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, tpl, curTplProp, "Start of circular reference, level " + ex.level.ToString(), null);
                        circularRef = null;
                    }
                }
                catch (Exception ex) { // All exceptions thrown by expression nodes are logged here.
                    Token t = VisEngine.curExprToken;
                    ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, tpl, curTplProp, 
                        (ex.InnerException != null ? ex.InnerException.Message : ex.Message), t, inst.Index());
                    continue;
                }
            }
            if (!(inst is VisForm)) 
                inst.AssignAndAdjustCanvas(inst.canvas);
        }

        public static FormTemplate LoadForm(string frmName) { // frmName without .vis
            //Debug.Print("LoadForm " + frmName);
            // Loads the vis-file into a frmTpl in the forms directory. This ensures that templates and properties can be addressed.
            // Compilation of the rows formulas isn't done. It is required for addressing database fields, but may cause error logging.
            FormTemplate frmTpl;
            List<FormTemplate> clones;
            int cloneID = -1;
            frmTpl = VisEngine.GetFormTemplate(frmName, ref cloneID); // Ensures that all form files are recorded.
            if (!Forms.TryGetValue(frmName, out clones)) return null; // The form doesn't exist as a file.
            frmTpl = VisTool.Parser.OpenVisFile(frmName); // Loads the vis-file into a form template.
            if (frmTpl == null) return null; // OpenVisFile has reported syntax errors.

            clones.Add(frmTpl);
            frmTpl.cloneID = 0; frmTpl.clones = clones; // Leave frmTpl.open as false.
            return frmTpl;
        }

        /* Copies all template properties from oldTpl to newTpl. This means creating new template properties. An alternative could be to let 
         * newTpl share oldTpl's properties (pointers to the same property objects). However, some expression nodes are clone dependent.
         * Examples: TemplateRefExpr (contains a reference to the target tplProp) and ControlJoinExpr (the ADOrelationID contains the cloneID).
         * This might be repaired so that the proper tplProp or ADOrelationID is computed at evaluation time. In that case, template props need no copy.
         */
        static void CopyTemplate(Template newTpl, Template oldTpl) {
            for (int i = 0; i < oldTpl.valPropArray.Length; i++)
                newTpl.valPropArray[i].Formula = oldTpl.valPropArray[i].Formula;
            newTpl.evtPropArray = oldTpl.evtPropArray; // They share it.
            for (int i = 0; i < oldTpl.evtPropArray.Length; i++)
                newTpl.evtPropArray[i].Formula = oldTpl.evtPropArray[i].Formula;
            foreach (var item in oldTpl.DesignProperties) {
                TplPropDesigner oldProp = item.Value;
                TplPropDesigner newProp = new TplPropDesigner(newTpl, oldProp.Name, typeof(object), PropertyKind.Designer);
                newProp.cmpPropIndex = newTpl.DesignProperties.Count;
                newProp.Formula = oldProp.Formula;
                newTpl.DesignProperties.Add(oldProp.Name, newProp);
            }
            newTpl.Name = oldTpl.Name;
            newTpl.Comment = oldTpl.Comment;
        }

        public static FormTemplate OpenForm(string frmName, Expression[] frmParams, out int cloneID) {
            //Debug.Print("OpenForm  " + frmName);
            // The form name is provided by the designer. To help him, .vis is removed when frmName ends with .vis. 
            // In case of errors, the return value is null and cloneID is -1.
            // Opens the vis file when necessary. Adds a form clone with the frmParameters. Reuses empty List clones.
            cloneID = -1;
            List<FormTemplate> clones;
            int newCloneID = -1;
            if (frmName == null || frmName.Length < 1) return null;
            if (frmName.EndsWith(".vis", StringComparison.OrdinalIgnoreCase))
                frmName = frmName.Remove(frmName.Length - 4);
            //sw1 = new Stopwatch(); sw2 = new Stopwatch(); count1 = 0;
            //sw1.Start();
            FormTemplate frmTpl = null;
            FormTemplate oldFrmTpl = VisEngine.GetFormTemplate(frmName, ref cloneID); // Ensures that all forms are recorded..
            if (cloneID < 0) return null; // The vis file doesn't exist.
            VisEngine.Forms.TryGetValue(frmName, out clones); // Will not fail because GetFormTemplate has checked.

            if (oldFrmTpl != null) { // Clone cloneID is open. Copy it to an empty slot.
                for (int i = 0; i < clones.Count; i++)  // Look for an empty slot.
                    if (clones[i] == null || !clones[i].open) { // An empty slot or a loaded form that isn't open.
                        newCloneID = i; break;
                    }
                if (newCloneID < 0) { // Create an empty slot.
                    newCloneID = clones.Count; clones.Add(null);
                }
                // Copy the clone to the empty slot.
                frmTpl = new FormTemplate(frmName);
                frmTpl.FormTpl = frmTpl;
                clones[newCloneID] = frmTpl;
                frmTpl.isDirty = oldFrmTpl.isDirty;
                foreach (var p in oldFrmTpl.Templates) {
                    Template tpl = frmTpl;
                    if (!(p.Value is FormTemplate)) { // The form template was added at create.
                        tpl = new Template(p.Key, p.Value.Info, frmTpl);
                        frmTpl.Templates.Add(p.Key, tpl);
                    }
                    CopyTemplate(tpl, p.Value);
                }

            } else { // There is no open clone.
                for (int i = 0; i < clones.Count; i++)  // First try to use a loaded form.
                    if (clones[i] != null && !clones[i].open) { // Open a loaded form that isn't open.
                        newCloneID = i; frmTpl = clones[i]; break;
                    }
                if (newCloneID < 0) { // The form must be loaded.
                    newCloneID = 0;
                    frmTpl = LoadForm(frmName); // Load and create the templates, add the form to clones.
                    if (frmTpl == null) return null;
                } // A new form has been created

            }
            frmTpl.cloneID = newCloneID; frmTpl.clones = clones; // Important for compilation of rows, etc.
            ResetTemplates(frmTpl); // In case the form is reused, delete all instances, reset Rows, etc.
            CompileRows(frmTpl); // Also needed when VisFile is just loaded by PathEngine to check a foreign form reference.
            CompileTemplates(frmTpl); // Necessary for the clone because some Expression nodes have a tpl reference or a cloneID, e.g. ADOrelationID.
            frmTpl.open = true;

            LockForm(frmTpl);
            frmTpl.exprParameters = frmParams;
            if (frmParams != null) {
                frmTpl.parameters = new object[frmParams.Length];
                for (int i = 0; i < frmParams.Length; i++)
                    frmTpl.parameters[i] = frmParams[i].Eval(VisEngine.curComponent, VisEngine.curTemplate);
            }
            frmTpl.callingContext = VisEngine.curComponent;
            frmTpl.RowBundle = new Bundle(1, null, frmTpl.propRows);

            VisEngine.refreshMode = VisEngine.RefreshMode.FormulaChange;
            //ResetUpdated(frmTpl);
            CreateAndCalculate(frmTpl);
            UnLockForm(frmTpl);
            VisEngine.refreshMode = VisEngine.RefreshMode.ValueChange; // All refresh done.
            //s5 = sw1.ElapsedMilliseconds;
            Show(frmTpl);
            frmTpl.RowBundle.Instances[0].cmp.GuiInstance.Focus(); // Make sure the form isn't hidden behind other forms.
            frmTpl.RowBundle.Instances[0].cmp.SetKeyFocus();

            //s6 = sw1.ElapsedMilliseconds;
            //Debug.Print(s2 + "\tOpen .vis " + frmName);
            //Debug.Print( (s3 - s2) + "\tRead vis-file");
            //Debug.Print((s4 - s3) + "\tTree + compile");
            //Debug.Print((s5 - s4) + "\tCreateInstanceTree");
            //Debug.Print(sw2.ElapsedMilliseconds + "\tSql queries: " + count1);
            //Debug.Print((s6 - s5)+"\tShow");
            //Debug.Print((sw1.ElapsedMilliseconds - s6) + "\tTime for 6*debug.print\n");
            cloneID = frmTpl.cloneID;
            return frmTpl;
        }

        public static void CloseAllFormClones(string frmName) { // Closes all the form bundles
            VisComponent oldCmp = curComponent; // Form close may call event handlers that evaluate stuff. Save state.
            Template oldTpl = curTemplate;
            TplProperty oldProp = curTplProp;
            Token oldToken = curExprToken;
            List<FormTemplate> clones;
            if (Forms.TryGetValue(frmName, out clones)) {
                for (int i = 0; i < clones.Count; i++) {
                    if (clones[i] != null && clones[i].RowBundle != null && clones[i].RowBundle.Instances.Count > 0) { // At least one Form to close
                        (clones[i].RowBundle.Instances[0].cmp.GuiInstance as Form).Close(); // Close a form. Will close all the forms in the bundle.
                        // Closing sets clones[i] = null. 
                    };
                }
            }
            curComponent = oldCmp; // Restore.
            curTemplate = oldTpl;
            curTplProp = oldProp;
            curExprToken = oldToken;
        }

        public static void CloseAllForms() { // Closes all forms and removes them from the Forms dictionary.
            // Currently used only by Studio.
            foreach (var frm in VisEngine.Forms) {
                CloseAllFormClones(frm.Key);
            }
            Forms.Clear();
            curComponent = null; // Avoid dangling components, etc.
            curTemplate = null;
            curTplProp = null;
            curExprToken = null;
        }

        public static void FormRefresh(FormTemplate frmTpl) { // Refreshes the form bundle and resets refreshMode.
            // If refreshMode is > Requery, refreshes  all open clones because they share the formulas.
            if (VisEngine.refreshMode <= RefreshMode.Requery) { // Refresh only this form bundle
                refreshOne(frmTpl); return;
            }
            // Formula, etc. changed. Refresh all open clones because they share the formulas.
            List<FormTemplate> clones = frmTpl.clones;
            for (int i = 0; i < clones.Count; i++)
                if (clones[i] != null && clones[i].open) refreshOne(clones[i]);
            if (!CompileOnly)
                VisEngine.refreshMode = VisEngine.RefreshMode.ValueChange; // All refresh done.
        }

        public static void refreshOne(FormTemplate frmTpl) { // Refreshes the form bundle. Doesn't reset refreshMode
            if (!LockForm(frmTpl))
                return; // May happen when an event handler is called during Refresh, e.g. FocusLost.
            //Debug.Print("Refresh: " + VisEngine.refreshMode);
            if (VisEngine.refreshMode >= RefreshMode.FormulaChange && !CompileOnly) {
                ResetTemplates(frmTpl); // Delete all instances, reset Rows, etc.
            }
            if (refreshMode >= RefreshMode.FormulaChange) {
                VisTool.ErrorLogging.ClearAllErrors(frmTpl.Name);
                CompileRows(frmTpl); // RefreshMode > FormulaChange done above.
                CompileTemplates(frmTpl);
            }
            //VisEngine.sw2.Restart();
            if (!CompileOnly) {
                ResetUpdated(frmTpl);
                CreateAndCalculate(frmTpl); // RefreshMode tells whether to requery or not.
            }
            UnLockForm(frmTpl); // Show triggers Paint which checks for update in progress.
            //VisEngine.sw2.Stop(); //VisEngine.sw1.Start();
            //Debug.Print(frmTpl.Name + ", createInstanceTree: " + VisEngine.sw2.ElapsedMilliseconds);
            //Debug.Print("Show: " + frmTpl.Name);

            Show(frmTpl);
            //sw1.Stop();
            //Debug.Print("Refresh values: " + sw1.ElapsedMilliseconds + " ms. Items: " + count1.ToString());
            //Debug.Print("Properties evaluated: " + count2 + "  Properties recalculated: " + count3);
            //Debug.Print("Controls added: "+ frmTpl.Name + ": "+ count1.ToString( ));
        }

        public static void MasterRefresh() {
            foreach (KeyValuePair<string, List<FormTemplate>> frm in Forms) {
                foreach (var cl in frm.Value)
                    if (cl != null && cl.open) refreshOne(cl); // A loaded form has no bundles.
            }
            if (!CompileOnly)
                VisEngine.refreshMode = VisEngine.RefreshMode.ValueChange; // All refresh done.
        }

        public static void ResetTemplates(FormTemplate frmTpl) {
            if (SystemItems.Popup != null) {
                SystemItems.Popup.Close(); SystemItems.Popup.Dispose();
            }
            // Let the form instances remain to avoid heavy flicker. Delete other components.
            foreach (KeyValuePair<string, Template> t in frmTpl.Templates) {
                Template tpl = t.Value; // Delete also Ignored templates. The Ignore field may just have been set.
                // Delete Find dictionaries:
                tpl.findProperties.Clear();
                foreach (var tplProp in tpl.DesignProperties) tplProp.Value.FindBundles = null;
                foreach (var tplProp in tpl.valPropArray) tplProp.FindBundles = null;

                if (tpl == frmTpl) { // Form template
                    if (frmTpl.RowBundle != null) // May be null when a form is reused and reset.
                    foreach (CmpProp cmpProp in frmTpl.RowBundle.Instances) {
                        VisComponent cmp = cmpProp.cmp;
                        cmp.childBundles = null; // Null to ensure that it will be reassigned at UpdateBundle. The number of child templates may change.
                    }
                    if (tpl.RowsExpr != null) tpl.RowsExpr.Unload(); // All rows will be loaded again and added to the adapter. Make sure the old rows are removed. SL 01-07-14.
                } else { // non-form templates
                    tpl.deleteInstances();
                }
                tpl.Updated = tpl.Visited = false; // Also reset in ResetUpdate, but may be needed for CompileTemplates
                tpl.RowsCompiled = false;
                tpl.RowsExpr = null; tpl.SQLstring = null;
            }
            //VisForm frm = frmTpl.bundles[0].Instances[0] as VisForm;
            //Form frmGui = frm.GuiInstance as Form;
            //if (frmGui.Controls.Count > 0) {
            //    Debug.Print(frmGui.Controls.Count + " Guis remain, e.g. " + frmGui.Controls[0].GetType().ToString());
            //}
            //if (frm.dictComponents.Count > 0) {
            //    Debug.Print(frm.dictComponents.Count + " cmps remain, e.g. " + frm.visComponents[0].GetType().ToString());
            //}

            //DataMaps[0].DataSet.Relations.Clear();
            //DataMaps[0].DataSet.Tables.Clear();
        }

        public static void ResetUpdated(FormTemplate frm) {
            ErrorLogging.ClearRuntimeErrors();
            foreach (KeyValuePair<string, Template> t in frm.Templates) { // Includes the form template itself
                Template tpl = t.Value;
                tpl.Updated = tpl.Visited = false;
                foreach (var tplProp in tpl.DesignProperties) 
                    tplProp.Value.FindBundles = null;
                foreach (var tplProp in tpl.valPropArray) 
                    tplProp.FindBundles = null;
                if (t.Value.Ignore) continue;

                if (tpl.RowBundle == null) continue; // may happen when reset is done several times.
                foreach (CmpProp cmpProp in tpl.RowBundle.Instances) { // Test for bundle == null if used for Forms
                    cmpProp.cmp.ResetUpdated(); // Reset internal dependencies.
                    if (cmpProp.cmp is ContainerComponent) {
                        (cmpProp.cmp as ContainerComponent).dictComponents.Clear();
                        // Components with Anonymous canvases, must clear them in ResetUpdated. E.g. PanelGrid.
                        //(c as ContainerComponent).GuiInstance.Controls.Clear(); // If cleared, all canvases disappear. Flicker !
                    }
                    cmpProp.cmp.canvas = null; // Tells assignCanvas that the component must be added to the canvas.
                    foreach (CmpProp prop in cmpProp.cmp.propertyValues)
                        if (prop.state != State.Initialized || refreshMode >= RefreshMode.FormulaChange) prop.state = State.Old;
                    foreach (CmpProp prop in cmpProp.cmp.designerPropValues) // At deletee desinger prop, prop may be null.
                        if (prop != null && prop.state != State.Initialized || refreshMode >= RefreshMode.FormulaChange)
                            prop.state = State.Old;
                }
            }
        }

        public static void CreateAndCalculate(FormTemplate frmTpl) {
            //Debug.Print("CreateAndCalculate " + tpl.Name);
            foreach (var item in frmTpl.Templates) { // The Form template must be the first for other templates to have a canvas.
                Template tpl = item.Value;
                if (tpl.Ignore) continue;
                CreateInstances(tpl); // might cross-reference
                VisComponent cmp = null;
                for (int index = 0; index < tpl.RowBundle.Instances.Count; index++) {
                    cmp = tpl.RowBundle.Instances[index].cmp;
                    CalculateProperties(tpl, cmp, index);
                    cmp.ComponentUpdated(); // Allow component to finish calculations before paint.
                    // cmp is assigned to canvas during evaluation of Canvas.
                }
                if (cmp != null) cmp.BundleUpdated();
                // Allows the component to make more complex calculations on the bundle, e.g. Trees or Connectors.
            }
        }

        public static void Show(FormTemplate frmtpl) {
            VisForm frm = frmtpl.RowBundle.Instances[0].cmp as VisForm;
            if (CompileOnly) {
                if (ShowForm != null)
                    ShowForm.Invoke(frm, new DesignModeEventArgs((Form)frm.GuiInstance, frmtpl));
                return;
            }
            if (!frm.GuiInstance.IsDisposed) { // May happen when designer deletes something. 
                frm.GuiInstance.Show(); // Wasn't necessary in earlier versions. Now needed.
                frm.GuiInstance.Refresh(); // Refresh Repaints the entire Form.
            }
            if (ShowForm != null) {
                ShowForm.Invoke(frm, new DesignModeEventArgs((Form)frm.GuiInstance, frmtpl));
                // Studio catches this event. It will usually refresh once again.
            }
        }

        public static void CreateInstances(Template tpl) {
            //Debug.Print("CreateInstances " + tpl.Name);
            ConstructorInfo CreateComponent;  // Reflection object to create controls.
            VisComponent dummyContext; // Used during evaluation of the Rows formula. Contains al the properties of the real components, but Row is null.
            // This allows references from the Rows formula to all kinds of properties as long as the property doesn't refer to the instance.

            if (tpl.Updated) return;
            CreateComponent = tpl.Info.Type.GetConstructor(new System.Type[] { typeof(int), typeof(DataRow), typeof(Template) });
            if (tpl.RowTpl != tpl) { // The rows are shared
                if (!tpl.RowTpl.Updated) // The main componenets aren't created.
                    CreateInstances(tpl.RowTpl); // Create the main components.
            } else { 

                // Create main components.
                if (tpl.RowsExpr == null) { NoRows(CreateComponent, tpl); return; }
                if (tpl.Visited) {
                    ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, tpl, VisEngine.curTplProp, "Cyclic reference when creating components", VisEngine.curExprToken, 0);
                    tpl.RowsExpr.RecordSet = null; NoRows(CreateComponent, tpl); return; 
                }
                tpl.Visited = true;
                dummyContext = CreateComponent.Invoke(new object[] { 0, null, tpl }) as VisComponent;
                dummyContext.tpl = tpl;

                // Calculate SQL
                curTplProp = tpl.propRows; // Prepare for errors.
                string sql = tpl.RowsExpr.Eval(dummyContext, tpl) as string; // Context needed for evaluation
                // Errors will return the SQL string as generated until the error. Will cause fault in Load.
                tpl.Visited = false;

                //Debug.Print("Query: Template: " + tpl.Name + " SQL: " + sql);
                if (sql != tpl.SQLstring || tpl.RowsExpr.createCount != tpl.createCount || refreshMode >= RefreshMode.Requery) { // Skip if SQL unchanged and no requery
                    tpl.SQLstring = sql; tpl.createCount = tpl.RowsExpr.createCount;
                    tpl.RowsExpr.Load(sql, tpl); // Retrieve data from database

                    //if (MaxRows < 1 && LimitedRows && tpl.RowsExpr.RecordSet.Rows.Count < 1) { // Show a dummy row.
                    //    System.Data.DataRow row = tpl.RowsExpr.RecordSet.NewRow();
                    //    tpl.RowsExpr.RecordSet.Rows.Add(row);
                    //}
                } // End requery
            }
            // Executed for Main as well as shared rows. Shared rows may experience faulty main rows.
            if (tpl.RowTpl.RowsExpr == null || tpl.RowTpl.RowsExpr.RecordSet == null) { // Error loading the data. Error reported by Load(sql).
                NoRows(CreateComponent, tpl); return;
            } 
            DataRow[] rows = tpl.RowTpl.RowsExpr.RecordSet.Rows.OfType<DataRow>().ToArray();
            if (tpl is FormTemplate && rows.Length == 0)
                NoRows(CreateComponent, tpl); // For a Form with zero rows, show one form anyway.
            else
                UpdateBundle(CreateComponent, rows, tpl);
            tpl.Updated = true; 
            return;
        }

        static void NoRows(ConstructorInfo ciCreateCtlInst, Template tpl) {
            UpdateBundle(ciCreateCtlInst, null, tpl); // Create bundle with one instance and no data row.
            tpl.Updated = true;
            tpl.RowTpl.addedRows.Clear();
        }

        static void UpdateBundle(ConstructorInfo ciCreateCtlInst, DataRow[] rows, Template tpl) {
            // Adjusts the length of tpl.RowBundle. Creates or deletes components accordingly.
            // Ensures that the row property of the new components refers to the Row bundle with a BundleRef, which provides the Index.
            try {
                if (tpl.RowBundle == null) tpl.RowBundle = new Bundle(rows == null ? 1 : rows.Length, null, tpl.propRows);
                int oldLength = tpl.RowBundle.Instances.Count;
                int newLength = (rows == null ? 1 : rows.Length);
                Bundle bundle = tpl.RowBundle;

                // Common part: The bundle exists in the right place with its old length. Add, remove and adjust components.
                // Keep them as far as possible. They may contain dialog data.

                if (oldLength < newLength)
                    bundle.Instances.Capacity = newLength;
                else if (oldLength > newLength) { // Remove instances
                    tpl.deleteInstances(bundle, newLength, oldLength - newLength);
                    bundle.Instances.RemoveRange(newLength, oldLength - newLength);
                }
                // When there are no rows, newLength is 1.
                int min = oldLength < newLength ? oldLength : newLength;
                for (int i = 0; i < min; i++) {
                    VisComponent cmp = bundle.Instances[i].cmp;
                    // ctrl.propertyValues.Length doesn't change. It is always as long as tpl.valPropArray.Length. Don't create it again.
                    if (cmp.designerPropValues.Length != tpl.DesignProperties.Count) { // A designer property has been added or deleted, e.g. by Studio.
                        cmp.designerPropValues = new CmpProp[tpl.DesignProperties.Count]; // Initialize all of them.
                        int dIndex = 0;
                        foreach (KeyValuePair<string, TplPropDesigner> tplProp in tpl.DesignProperties) {
                            cmp.designerPropValues[dIndex++] = new CmpProp(tplProp.Value, cmp);
                        }
                    }
                    if (rows == null)  // There are no Rows
                        cmp.Row = null; // Might have been rows before. 
                    else {
                        cmp.Row = rows[i]; // Assign the new row to the old components.
                        //if (tpl.TargetTpl) rows[i][ctrl.Row.Table.Columns.Count - 1] = bundle.Instances[i];
                        // Make the last field (Column) refer to the component. See TargetTpl and RecordSetExpr, addCtrlField.
                    }
                }

                for (int i = oldLength; i < newLength; i++) {  // Create the new Components by reflection   
                    VisComponent viscomp;
                    viscomp = ciCreateCtlInst.Invoke(new object[] { tpl.findProperties.Count, rows == null ? null : rows[i], tpl }) as VisComponent;
                    if (viscomp.GuiInstance != null) viscomp.GuiInstance.Tag = viscomp;
                    CmpProp cmpProp = viscomp.propertyValues[VisComponent.RowsID];
                    cmpProp.Value = new BundleRef(tpl.RowBundle, i, null); // Address the rows bundle in the same way as other bundles.
                    bundle.Instances.Add(cmpProp);
                }
            }
            catch (Exception ex) {
                Debug.Print("Exception during UpdateBundle: " + tpl.Name + "  " + ex.Message);
                return;
            }
        }
    }

    //-------------------------------------------------------------------------
    public class ComponentInfo {
        public Type Type;
        public VisClassAtt ClassAttribute;
        public Dictionary<string, PropInfo> PropertyByName = new Dictionary<string, PropInfo>(50, StringComparer.OrdinalIgnoreCase);
        // All properties, including functions and events. Excluding Designer properties.
        public Dictionary<string, ValueInfo> ValuePropertyByName = new Dictionary<string, ValueInfo>(30, StringComparer.OrdinalIgnoreCase);
        public Dictionary<string, FunctInfo> FunctionByName = new Dictionary<string, FunctInfo>(10, StringComparer.OrdinalIgnoreCase);
        public Dictionary<string, EvtInfo> EventByName = new Dictionary<string, EvtInfo>(30, StringComparer.OrdinalIgnoreCase);

        public ComponentInfo(Type type, VisClassAtt classAtt) {
            this.Type = type;
            this.ClassAttribute = classAtt;
        }

        public static void ReflectDLLs() { // Looks at all files in the program directory and extracts component and System attributes.
            Stopwatch sw1 = new Stopwatch();
            int count1 = 0;
            sw1.Start();

            try {
                var files = Directory.GetFiles(VisEngine.ProgramDirectory);
                foreach (var file in files) {
                    if (!Path.GetExtension(file).ToUpper().EndsWith("DLL")) continue;
                    var assembly = Assembly.LoadFrom(file);

                    // Get all Components and System items
                    foreach (Type type in assembly.GetTypes()) {
                        if (type.IsSubclassOf(typeof(VisComponent))) { // Component info
                            CreateComponentInfo(type, ref count1);
                        } // End Get component.

                     // Get all system items, including methods, enum methods, and fields.
                     else if (type == typeof(SystemItems)) { // type.IsClass && type.Name=="SystemItems"
                            foreach (MethodInfo m in type.GetMethods()) { // BindingFlags.Public return no methods.
                                var attributes = m.GetCustomAttributes(typeof(VisSystemAtt), true);
                                if (attributes.Length > 0) {
                                    var s = new SystemMethodInfo(m, (VisSystemAtt)attributes[0]);
                                    if (VisEngine.SystemMethodByName.ContainsKey(m.Name)) {
                                        Debug.Print("System method " + m.Name + " exists twice"); // Overloading not supported.
                                        continue;
                                    }
                                    VisEngine.SystemMethodByName.Add(m.Name, s); continue;
                                };
                            }
                            foreach (FieldInfo f in type.GetFields()) {
                                var attributes = f.GetCustomAttributes(typeof(VisSystemAtt), true);
                                if (attributes.Length > 0) {
                                    SystemFieldInfo s = new SystemFieldInfo(f, (VisSystemAtt)attributes[0]);
                                    VisEngine.SystemFieldByName.Add(f.Name, s); continue;
                                };
                            };
                        }
                    } // End for each type
                } // End for each file
            }
            catch (Exception ex) {
                MessageBox.Show("Unexpected exception in ReflectDLLs: " + ex.Message);
                Debug.Print("Unexpected exception in ReflectDLLs: " + ex.Message);
            }
            sw1.Stop();
            Debug.Print("ReflectDLLs: " + sw1.ElapsedMilliseconds + " ms. Items: " + count1.ToString());
        }

        static ComponentInfo CreateComponentInfo(Type type, ref int count1) {
            ComponentInfo componentInfo;
            if (VisEngine.ComponentByName.TryGetValue(type.Name, out componentInfo))
                return componentInfo; // The component is a base component and has been created already.
            var classAtt = (VisClassAtt)Attribute.GetCustomAttribute(type, typeof(VisClassAtt));
            componentInfo = new ComponentInfo(type, classAtt);

            if (type != typeof(VisComponent)) { // Copy all properties from base type.
                Type basetype = type.BaseType;
                ComponentInfo baseInfo;
                if (!VisEngine.ComponentByName.TryGetValue(basetype.Name, out baseInfo))
                    // Base-component info not created. Create it.
                    baseInfo = CreateComponentInfo(basetype, ref count1);
                foreach (var pi in baseInfo.PropertyByName)
                    componentInfo.PropertyByName.Add(pi.Key, pi.Value);
                foreach (var pi in baseInfo.ValuePropertyByName)
                    componentInfo.ValuePropertyByName.Add(pi.Key, pi.Value);
                foreach (var pi in baseInfo.EventByName)
                    componentInfo.EventByName.Add(pi.Key, pi.Value);
                foreach (var pi in baseInfo.FunctionByName)
                    componentInfo.FunctionByName.Add(pi.Key, pi.Value);
            }

            // Get all value properties, e.g. Top and Rows
            var properties = type.GetProperties(BindingFlags.Instance | BindingFlags.Public).ToList();
            foreach (var propInfo in properties) {
                var attributes = propInfo.GetCustomAttributes(typeof(VisPropAtt), true);
                if (attributes.Length == 0) continue;
                VisPropAtt v = (VisPropAtt)attributes[0];
                if (!v.Exposed) continue;
                ValueInfo newInfo = new ValueInfo(propInfo, v, componentInfo.ValuePropertyByName.Count);
                ValueInfo oldInf;
                if (componentInfo.ValuePropertyByName.TryGetValue(propInfo.Name, out oldInf)) { // Replace the old ValueInfo
                    newInfo.propIndex = oldInf.propIndex;
                    componentInfo.ValuePropertyByName[propInfo.Name] = newInfo;
                    componentInfo.PropertyByName[propInfo.Name] = newInfo;
                } else { // Add the new ValueInfo.
                    componentInfo.PropertyByName.Add(propInfo.Name, newInfo);
                    componentInfo.ValuePropertyByName.Add(propInfo.Name, newInfo);
                }
                FieldInfo pi = type.GetField(propInfo.Name + "ID", BindingFlags.Static | BindingFlags.Public);
                // Let the property address its own formula and value by an index.
                if (pi != null) {
                    pi.SetValue(type, newInfo.propIndex);
                }
                count1++;
            }

            // Get all function properties, e.g. HPos( ), and Event handlers, e.g. Click.
            var functions = type.GetMethods(BindingFlags.Instance | BindingFlags.Public).ToList();
            foreach (var propertyInfo in functions) {
                var attributes = propertyInfo.GetCustomAttributes(typeof(VisPropAtt), true);
                if (attributes.Length == 0) continue; // Only used when there is a VisPropAtt.
                VisPropAtt p = (VisPropAtt)attributes[0];
                if (p.PropType == VisPropAtt.PropertyType.Eventhandler) {
                    if (!componentInfo.EventByName.ContainsKey(propertyInfo.Name)) {
                        var eventInfo = new EvtInfo(propertyInfo, p, componentInfo.EventByName.Count);
                        componentInfo.PropertyByName.Add(propertyInfo.Name, eventInfo);
                        componentInfo.EventByName.Add(propertyInfo.Name, eventInfo);
                    }
                } else {
                    if ( !componentInfo.FunctionByName.ContainsKey(propertyInfo.Name)) {
                        var functInfo = new FunctInfo(propertyInfo, p, componentInfo.FunctionByName.Count);
                        componentInfo.PropertyByName.Add(functInfo.Name, functInfo);
                        componentInfo.FunctionByName.Add(functInfo.Name, functInfo);
                    }
                }
                count1++;
            }

            VisEngine.ComponentByName.Add(type.Name, componentInfo);
            return componentInfo;
        }
    }

    //---------------------------------------------------------------------------------------
    public class PropInfo { // Includes value properties, events and functions, but not desiger properties.
        // Value properties have a formula and a value, Events have only a formula. Functions have neither formula, nor value.
        // Designer properties have a formula and a value.
        public string Name;
        public VisPropAtt VisAttribute;
        public int propIndex; // Index 0 to FirstEventInfo-1 are for value properties. FirstEventInfo and on are for event info.
        // Value properties have a counterpart in InstProperties.
    }

    //---------------------------------------------------------------------------------------
    public class ValueInfo : PropInfo {
        public Type Type;
        public PropertyInfo PropertySetter;

        public ValueInfo(PropertyInfo propertyInfo, VisPropAtt attr, int propIndex) {
            this.Name = propertyInfo.Name;
            this.PropertySetter = propertyInfo;
            this.Type = propertyInfo.PropertyType;
            VisAttribute = attr;
            this.propIndex = propIndex;
        }
    }

    //---------------------------------------------------------------------------------------
    public class FunctInfo : PropInfo {
        public ParameterInfo[] Parameters;
        public MethodInfo PropertySetter;

        public FunctInfo(MethodInfo methodInfo, VisPropAtt attr, int propIndex) {
            this.Name = methodInfo.Name;
            this.PropertySetter = methodInfo;
            this.Parameters = methodInfo.GetParameters();
            VisAttribute = attr;
            this.propIndex = propIndex;
        }
    }

    //---------------------------------------------------------------------------------------
    public class EvtInfo : PropInfo { // Event properties are like function properties, but have a formula
        public MethodInfo PropertySetter;

        public EvtInfo(MethodInfo eventInfo, VisPropAtt attr, int propIndex) {
            this.Name = eventInfo.Name;
            this.PropertySetter = eventInfo;
            VisAttribute = attr;
            this.propIndex = propIndex;
        }
    }

    //---------------------------------------------------------------------------------------
    public class SystemMethodInfo : PropInfo {
        public VisSystemAtt attr;
        public ParameterInfo[] Parameters;
        public MethodInfo PropertySetter;

        public SystemMethodInfo(MethodInfo methodInfo, VisSystemAtt attr) {
            this.Name = methodInfo.Name;
            this.PropertySetter = methodInfo;
            this.Parameters = methodInfo.GetParameters();
            this.attr = attr;
        }
    }

    //---------------------------------------------------------------------------------------
    public class SystemFieldInfo : PropInfo {
        public VisSystemAtt attr;
        public FieldInfo PropertySetter;

        public SystemFieldInfo(FieldInfo fieldInfo, VisSystemAtt attr) {
            this.Name = fieldInfo.Name;
            this.PropertySetter = fieldInfo;
            this.attr = attr;
        }
    }

    //---------------------------------------------------------------------------------------
    public class DesignModeEventArgs : EventArgs {
        public Form form;
        public FormTemplate formTempl;

        public DesignModeEventArgs(Form form, FormTemplate frmTemp) {
            this.form = form;
            this.formTempl = frmTemp;
        }
    }

}
