﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Windows.Forms;
using VisTool;
using uVisStudio.VisToolbox;

namespace uVisStudio.VisDesigner {

    public class HookState {
        public virtual void MouseLeftButtonDown(Point point) { Hook.Handled = false; return; }
        public virtual void MouseMove(Point point) { Hook.Handled = false; return; }
        public virtual void MouseLeftButtonUp(Point point) { Hook.Handled = false; return; }
        public virtual void KeyDown(Keys key) { Hook.Handled = false; return; }
        public virtual void KeyUp(Keys key) { Hook.Handled = false; return; }

        // Static methods ---------------------------------------------------
        public static bool IsDesignMode() {
            return VisEngine.InteractionMode != VisEngine.interactionMode.AsEndUser;
        }

        public static bool IsDesignClick() {
            var isCtrl = VisEngine.InteractionMode == VisEngine.interactionMode.DesignWithControl;
            var isAlt = VisEngine.InteractionMode == VisEngine.interactionMode.DesignWithAlt;
            var isDesigner = VisEngine.InteractionMode == VisEngine.interactionMode.AsDesigner;
            bool result = Control.ModifierKeys == System.Windows.Forms.Keys.Control && isCtrl ||
                Control.ModifierKeys == System.Windows.Forms.Keys.Alt && isAlt || isDesigner;
            return result;
        }

        public static ContainerComponent FindContainer(ContainerComponent container, Point screenPoint) {
            if (container == null) return null;
            for (int i = 0; i < container.visComponents.Length; i++) { // Add components in increasing zOrder.
                ContainerComponent ci = container.visComponents[i] as ContainerComponent;
                if (ci != null) {
                    Point ciOnScreen = ci.GuiInstance.PointToScreen(new Point(0, 0));
                    if (ciOnScreen.X <= screenPoint.X && ciOnScreen.X + VisComponent.PhyS(ci.width) >= screenPoint.X
                        && ciOnScreen.Y <= screenPoint.Y && ciOnScreen.Y + VisComponent.PhyS(ci.height) >= screenPoint.Y)
                        return FindContainer(ci, screenPoint);
                }
            }
            return container; // Nothing inside the container matched.
        }

        public static string CreateTemplateName(FormTemplate frm, string typeName) {
            int number = 0;
            while (true) {
                number++;
               Template template;
                var newTemplateName = typeName + number;
                if (frm.DesignProperties.ContainsKey(newTemplateName)) continue; // May not conflict with a form property name.
                if (!frm.Templates.TryGetValue(newTemplateName, out template)) return newTemplateName;
            }
        }

        public static void SortComponentsBySize(ref List<VisComponent> Components) {
            if (Components.Count > 1)
                SortShapes(0, Components.Count - 1, ref Components);
            List<VisComponent> newComponents = new List<VisComponent>();
            for (int i = Components.Count - 1; i >= 0; i--) {
                if (!newComponents.Contains(Components[i]))
                    newComponents.Add(Components[i]);
            }
            Components = newComponents;
        }

        public static void SortShapes(int left, int right, ref List<VisComponent> Components) {
            int i = left, j = right;
            VisComponent pivot = Components[(left + right) / 2];
            while (i <= j) {
                while (Components[i].BiggerSize(Components[i], pivot)) i++;
                while (Components[j].BiggerSize(pivot, Components[j])) j--;
                if (i <= j) {
                    VisComponent tmp = Components[i];
                    Components[i] = Components[j];
                    Components[j] = tmp;
                    i++; j--;
                }
            }
            // Recursive calls
            if (left < i - 1)
                SortShapes(left, i - 1, ref Components);
            if (i < right)
                SortShapes(i, right, ref Components);
        }

        public static void DesignClick(Form form, Point point) { // Form must be non-null.
            var clickedCmps = Hook.GetClickedInstances(form, point);
            if (clickedCmps.Count == 0) return;
            SortComponentsBySize(ref clickedCmps);

            int last = SelectionBox.clickedStack.Count - 1;
            VisComponent newClicked = clickedCmps[clickedCmps.Count - 1];

            if (last >= 0 && SelectionBox.clickedStack[last] == newClicked) { // Click in the existing clicked list. Continue in the list
                SelectionBox.latestIndex--;
                if (SelectionBox.latestIndex < 0) SelectionBox.latestIndex = last;
                newClicked = SelectionBox.clickedStack[SelectionBox.latestIndex];
            } else {
                SelectionBox.clickedStack = clickedCmps;
                SelectionBox.latestIndex = clickedCmps.Count - 1;
            }
            var selEnum = SelectionBox.selections.GetEnumerator();
            if (selEnum.MoveNext() && selEnum.Current.Value.cmp.canvas != newClicked.canvas)
                return; // Trying to add a selection with another canvas than the present selections. Not allowed.
            SelectionBox.selected = new SelectionBox(newClicked);
            TemplateHandler.SelectTemplate(newClicked); // Triggers Studio to update Property grids.
        }

        public static List<MoveTransaction> GetMoves(ContainerComponent container, VisForm newFrm) {
            ScrollableControl target = null;
            FormTemplate newFrmTpl = null;
            if (container != null) target = container.GuiInstance as ScrollableControl;
            var moveTransactions = new List<MoveTransaction>();
            if (newFrm != null) newFrmTpl = newFrm.tpl as FormTemplate;

            foreach (var kv in SelectionBox.selections) {
                SelectionBox sel = kv.Value;
                MoveTransaction moveTrans = new MoveTransaction(sel.cmp, newFrmTpl);
                if (newFrm != null) {
                    moveTrans.CanvasFormula = container.GetCanvasFormula();
                    Point screenPoint = sel.blueBox.PointToScreen(new Point(0, 0)); // New screen coordinates of component's top-left.
                    Point relativePoint = target.PointToClient(screenPoint); // Relative position in target.
                    moveTrans.NewLeft = (int)((relativePoint.X + target.HorizontalScroll.Value) / SystemItems.Zoom2 + 0.5);
                    moveTrans.NewTop = (int)((relativePoint.Y + target.VerticalScroll.Value) / SystemItems.Zoom2 + 0.5);
                    // Can not use container.Logical( ) because the component is not yet in the container. 
                } else { // Moving a form.
                    moveTrans.NewLeft = sel.blueBox.Left;
                    moveTrans.NewTop = sel.blueBox.Top;
                }
                moveTransactions.Add(moveTrans);
            }
            return moveTransactions;
        }

        public static void ResizeInstances(HookResizeMode resizeMode, Point point) {
            int xDif = (int)(point.X - Hook.StartPoint.X);
            int yDif = (int)(point.Y - Hook.StartPoint.Y);

            foreach (var kv in SelectionBox.selections) {
                SelectionBox sel = kv.Value;
                VisComponent cmp = sel.cmp;
                Point scrPos;
                int width = cmp is VisForm ? (cmp as VisForm).form.ClientSize.Width : VisComponent.PhyS(cmp.width);
                int height = cmp is VisForm ? (cmp as VisForm).form.ClientSize.Height : VisComponent.PhyS(cmp.height);
                var box = sel.blueBox;

                if (resizeMode == HookResizeMode.West) {
                    if (xDif == 0) return;
                    if (xDif >= box.Width) continue;
                    scrPos = cmp.PointToScreen(new Point(xDif, 0));
                    box.Location = scrPos;
                    box.Width = width - xDif;
                }
                if (resizeMode == HookResizeMode.North) {
                    if (yDif == 0) return;
                    if (yDif >= box.Height) continue;
                    scrPos = cmp.PointToScreen(new Point(0, yDif));
                    box.Location = scrPos;
                    box.Height = height - yDif;
                }
                if (resizeMode == HookResizeMode.East) {
                    if (xDif == 0) return;
                    box.Width = width + xDif;
                }
                if (resizeMode == HookResizeMode.South) {
                    if (yDif == 0) return;
                    box.Height = height + yDif;
                }
                box.Invalidate(); // Otherwise we see old blue boxes.
            }
        }

        public static void ResizeInstancesUpdate() {
            if (Hook.mouseCapture != null) {
                Hook.mouseCapture.Capture = false;
                Hook.mouseCapture = null;
            }

            var resizeTransactions = new List<ResizeTransaction>();

            foreach (var kv in SelectionBox.selections) {
                SelectionBox sel = kv.Value;
                var resizeTrans = new ResizeTransaction(sel.cmp);
                if (sel.cmp.canvas == null) { // Resizing a VisForm.
                    resizeTrans.NewLeft = VisComponent.LogS(sel.blueBox.Left - sel.relativeX); // -4
                    resizeTrans.NewTop = VisComponent.LogS(sel.blueBox.Top - sel.relativeY); // -23
                    resizeTrans.NewWidth = VisComponent.LogS(sel.blueBox.Width + 2 * sel.relativeX);
                    resizeTrans.NewHeight = sel.blueBox.Height + sel.relativeX + sel.relativeY;
                } else {
                    ScrollableControl target = sel.cmp.canvas.GuiInstance as ScrollableControl;
                    Point screenPoint = sel.blueBox.PointToScreen(new Point(0, 0)); 
                    Point relativePoint = target.PointToClient(screenPoint); // Relative position in target.
                    resizeTrans.NewLeft = VisComponent.LogS(relativePoint.X + target.HorizontalScroll.Value + sel.blueBox.Htrim); // Logical position in target.
                    resizeTrans.NewTop = VisComponent.LogS(relativePoint.Y + target.VerticalScroll.Value + sel.blueBox.Vtrim);
                    resizeTrans.NewWidth = VisComponent.LogS(sel.blueBox.Width - sel.blueBox.Htrim * 2);
                    resizeTrans.NewHeight = VisComponent.LogS(sel.blueBox.Height - sel.blueBox.Vtrim * 2);
                }
                resizeTransactions.Add(resizeTrans);
            }
            if (resizeTransactions.Count > 0) {
                Controller.Instance.ResizeTemplates(resizeTransactions);
            }
        }

        public static void DeleteSelected() {
            var templates = new List<Template>();
            foreach (var sel in SelectionBox.selections) templates.Add(sel.Value.tpl);
            Controller.Instance.DeleteTemplates(templates);
        }

        public enum StepDirection { // For moving components pixel by pixel with the arrow keys.
            Left,
            Right,
            Up,
            Down
        }

        public static void MoveOneStep(StepDirection direction) { // Move components pixel by pixel with the arrow keys.
            var moveTransactions = new List<MoveTransaction>();
            foreach (var kv in SelectionBox.selections) {
                VisComponent cmp = kv.Value.cmp;
                var moveTrans = new MoveTransaction(cmp, cmp.tpl.FormTpl);
                moveTrans.NewLeft = cmp.left;
                moveTrans.NewTop = cmp.top;

                if (direction == StepDirection.Left) {
                    moveTrans.NewLeft = cmp.Left - 1;
                } else if (direction == StepDirection.Right) {
                    moveTrans.NewLeft = cmp.Left + 1;
                } else if (direction == StepDirection.Up) {
                    moveTrans.NewTop = cmp.Top - 1;
                } else if (direction == StepDirection.Down) {
                    moveTrans.NewTop = cmp.Top + 1;
                }
                moveTransactions.Add(moveTrans);
            }

            if (moveTransactions.Count > 0) {
                Controller.Instance.MoveTemplates(moveTransactions);
            }
        }
    }

    //----------------------------------------------------------------------------------------
    public class HookNeutralState : HookState {
        private bool testMoveState = false;
        SelectionBox clicked;

        public override void MouseLeftButtonDown(Point point) {
            if (VisEngine.InteractionMode == VisEngine.interactionMode.AsEndUser) { Hook.Handled = false; return; }
            Hook.currentForm = Hook.GetClickedForm(point);

            if (VisToolboxItem.SelectedTool != null) { // Tool selected. Create component.
                if (VisToolboxItem.SelectedTool.ComponentInfo.Type == typeof(VisForm) || Hook.currentForm != null) {
                    SelectionBox.DeleteAll();
                    Hook.State = new HookCreateState(point); return;
                }
            }

            if (Hook.currentForm == null) { Hook.Handled = false; return; } // Click outside VisForms.
            clicked = SelectionBox.SelectionClicked(point);
            if (clicked != null) { // An existing blue box was clicked.
                if (IsDesignClick()) { // Delete all boxes and select one, maybe higher or lower in the stack.
                    SelectionBox.DeleteAll(); DesignClick(Hook.currentForm, point);
                    return;
                }
                if (Hook.IsKeyPressed(Keys.ShiftKey)) { // Deselect clicked box.
                    SelectionBox.DeleteBox(clicked.tpl); // Updates selected too.
                    return;
                }
                // Prepare move or resize of the clicked box.
                var resizeMode = clicked.ResizeMode(point);
                if (resizeMode != HookResizeMode.Nothing) { // Click on the border.
                    Hook.State = new HookResizeState(resizeMode);
                    SelectionBox.selected = clicked;
                    if (resizeMode == HookResizeMode.West || resizeMode == HookResizeMode.East)
                        Cursor.Current = Cursors.VSplit;
                    else
                        Cursor.Current = Cursors.HSplit;
                    SelectionBox.selected.form.Capture = true;
                    Hook.mouseCapture = SelectionBox.selected.form;
                } else
                    testMoveState = true;
                return;
            }

            // Click outside a blue box.
            if (IsDesignClick()) { // Select this one, deselect others. 
                SelectionBox.DeleteAll(); DesignClick(Hook.currentForm, point);
                // Don't set testMoveState. It will cause a selected form to be deselected at MouseUp.
                return;
            }
            if (SelectionBox.selections.Count > 0 && Hook.IsKeyPressed(Keys.ShiftKey)) { // Add selection.
                DesignClick(Hook.currentForm, point);
                return;
            }
            if (SelectionBox.selections.Count > 0) { // Deselect all.
                SelectionBox.DeleteAll(); return; 
            }

            Hook.Handled = false;
            return;
        }

        public override void MouseMove(Point point) {
            if (VisEngine.InteractionMode == VisEngine.interactionMode.AsEndUser) { Hook.Handled = false; return; }
            if (VisToolboxItem.SelectedTool != null && IsDesignMode()) { // Tool selected.
                Hook.currentForm = Hook.GetClickedForm(point);
                if (VisToolboxItem.SelectedTool.ComponentInfo.Type == typeof(VisForm)) // Create VisForm.
                    Cursor.Current = HookMoveState.cForm;
                else { // Create other component.
                    if (Hook.currentForm == null) return;
                    Cursor.Current = Cursors.Cross;
                }
                return;
            }

            clicked = SelectionBox.SelectionClicked(point);
            if (clicked != null) SelectionBox.selected = clicked;

            SelectionBox box = SelectionBox.selected;
            VisComponent c;

            if (box == null || (c = box.cmp) == null) { Hook.Handled = false; return; }
            if (testMoveState) {
                var distSquare = Math.Pow(Hook.StartPoint.X - point.X, 2) + Math.Pow(Hook.StartPoint.Y - point.Y, 2);

                if (distSquare >= 9) { // Move at least 3 pixels.
                    Hook.State = new HookMoveState();
                    testMoveState = false;
                    if (SelectionBox.selected == null) return;
                    SelectionBox.selected.form.Capture = true;
                    Hook.mouseCapture = SelectionBox.selected.form;
                    Point screenLoc = c.PointToScreen(new Point(0, 0)); // For a Form, it is the topLeft of the client area.
                    SelectionBox.dragOffsetX = point.X - screenLoc.X;
                    SelectionBox.dragOffsetY = point.Y - screenLoc.Y;
                    return;
                }
            }

            HookResizeMode mode = box.ResizeMode(point);
            if (mode != HookResizeMode.Nothing) { // Is close to a border.
                if (mode == HookResizeMode.West || mode == HookResizeMode.East)
                    Cursor.Current = Cursors.VSplit;
                else
                    Cursor.Current = Cursors.HSplit;
                // Don't capture mouse at move, only at click.
                return;
            } 
            if (box.IsMoveable(point)) {
                Cursor.Current = Cursors.Hand;
                return;
            }
            Cursor.Current = Cursors.Default;
            Hook.Handled = false;
        }

        public override void MouseLeftButtonUp(Point point) {
            if (VisEngine.InteractionMode != VisEngine.interactionMode.AsEndUser) {
                clicked = SelectionBox.SelectionClicked(point);
                if (testMoveState && clicked != null && clicked.cmp is VisForm)
                    SelectionBox.DeleteAll(); // Simple click on a form.
                testMoveState = false;  // Didn't move.
                if (clicked != null) return; // Stop click after design select.
                // Click events occur at button up, design click at button down.
            }
            Hook.Handled = false;
        }

        public override void KeyDown(Keys key) {
            if (SelectionBox.selected != null && SelectionBox.selected.frm.GuiInstance.Focused) {
                // Checked that the form has focus, not property gridline, etc.
                if (Keys.Left == key) HookState.MoveOneStep(HookState.StepDirection.Left);
                else if (Keys.Right == key) HookState.MoveOneStep(HookState.StepDirection.Right);
                else if (Keys.Up == key) HookState.MoveOneStep(HookState.StepDirection.Up);
                else if (Keys.Down == key) HookState.MoveOneStep(HookState.StepDirection.Down);
                else if (Keys.Delete == key &&
                    System.Windows.Input.Keyboard.Modifiers == (System.Windows.Input.ModifierKeys.Control)) DeleteSelected();
                else Hook.Handled = false; // Takes care of e.g. Ctrl down.
            } else
                Hook.Handled = false; return;
        }
    }

    //----------------------------------------------------------------------------------------
    public class HookMoveState : HookState {
        public static Cursor cCopy = new Cursor(VisEngine.ProgramDirectory + "\\Icons\\compCopy.cur");
        public static Cursor cForm = new Cursor(VisEngine.ProgramDirectory + "\\Icons\\form.cur");
        Form form;

        public override void MouseMove(Point point) { // Move boxes.
            form = Hook.GetClickedForm(point);
            if (Hook.IsKeyPressed(Keys.ControlKey) && !(SelectionBox.selected.cmp is VisForm) || form != null && form != SelectionBox.selected.form)
                Cursor.Current = cCopy;  // Copy when Ctrl down for a non-form component or when moving to another Form.
            else
                Cursor.Current = Cursors.Hand;
            int xDif = point.X - SelectionBox.dragOffsetX - SelectionBox.selected.blueBox.Left;
            int yDif = point.Y - SelectionBox.dragOffsetY - SelectionBox.selected.blueBox.Top;
            foreach (var kv in SelectionBox.selections) {
                BlueBox blueBox = kv.Value.blueBox;
                blueBox.Left += xDif; // Physical pixels.
                blueBox.Top += yDif;
                blueBox.Invalidate();
            }
        }

        public override void KeyDown(Keys key) {
            if (Hook.IsKeyPressed(Keys.ControlKey) && !(SelectionBox.selected.cmp is VisForm) || form != null && form != SelectionBox.selected.form)
                Cursor.Current = cCopy;
            Hook.Handled = false;
        }

        public override void KeyUp(Keys key) {
            if (! Hook.IsKeyPressed(Keys.ControlKey) && form == SelectionBox.selected.form)
                Cursor.Current = Cursors.Hand; // Move only within same Form.
            Hook.Handled = false;
        }

        public override void MouseLeftButtonUp(Point point) { // point is screen coordinates of the dragged point.
            ContainerComponent targetContainer = null; // null when moving a form.
            VisForm frm = null; // null when moving a form. 
            if (!(SelectionBox.selected.cmp is VisForm)) { // Not moving a form.
                form = Hook.GetClickedForm(point); // May be in another visForm.
                if (form == null) { // Moving outside the visible part of a canvas. Keep the canvas.
                    targetContainer = SelectionBox.selected.cmp.canvas;
                    frm = SelectionBox.selected.frm;
                } else {
                    frm = form.Tag as VisForm; // Target form.
                    targetContainer = FindContainer(frm, point); // May find a canvas on the form.
                }
            }
            List<MoveTransaction> moves = GetMoves(targetContainer, frm); // Transform selections to ... 
            foreach (MoveTransaction source in moves) {
                if (source.Instance == targetContainer) { // The target container is in the move list.
                    targetContainer = targetContainer.canvas; // Choose the surrounding container.
                    // All selections are on the same canvas, so the surrounding container is not in selections.
                    moves = GetMoves(targetContainer, frm);
                    break;
                }
            }
            Hook.State = new HookNeutralState(); // The change commands below may generate MsgBoxes. Allow simple cursor move.
            if (Hook.IsKeyPressed(Keys.ControlKey) && !(SelectionBox.selected.cmp is VisForm) || form != null && form != SelectionBox.selected.form)
                Controller.Instance.CopyTemplates(moves);
            else
                Controller.Instance.MoveTemplates(moves);
        }
    }

    //----------------------------------------------------------------------------------------
    public class HookResizeState : HookState {
        private HookResizeMode hookResizeMode;

        public HookResizeState(HookResizeMode hookResizeMode) {
            this.hookResizeMode = hookResizeMode;
        }

        public override void MouseMove(Point point) {
            ResizeInstances(hookResizeMode, point); // Embed here. Used only from here.
        }

        public override void MouseLeftButtonUp(Point point) {
            Hook.State = new HookNeutralState();  // The resize command below may generate MsgBoxes. Allow simple cursor move.
            ResizeInstancesUpdate(); // May refresh, etc. Embed here. Used only from here.
        }
    }

    //----------------------------------------------------------------------------------------
    public class HookCreateState : HookState {
        private Controller controller = Controller.Instance;

        public HookCreateState(Point point) {
            SelectionBox.DrawingBox = new BlueBox(point, new Size(1, 1));
        }

        public override void MouseMove(Point point) {
            Hook.currentForm = Hook.GetClickedForm(Hook.StartPoint);
            int x = Hook.StartPoint.X < point.X ? Hook.StartPoint.X : point.X;
            int y = Hook.StartPoint.Y < point.Y ? Hook.StartPoint.Y : point.Y;
            int width = Math.Abs(point.X - Hook.StartPoint.X);
            int height = Math.Abs(point.Y - Hook.StartPoint.Y);
            if (width > 0 || height > 0) {
                SelectionBox.DrawingBox.Size = new Size(width, height);
                //Debug.Print("CreateState.MoseMove " + width + ", " + height);
                SelectionBox.DrawingBox.Invalidate(); // Otherwise we see old blue boxes.
            }
        }

        public override void MouseLeftButtonUp(Point point) {
            ContainerComponent container = null;
            Hook.currentForm = Hook.GetClickedForm(Hook.StartPoint);
            if (Hook.currentForm != null) // Dropping a VisForm may be on Studio, which isn't a VisForm.
                container = Hook.currentForm.Tag as ContainerComponent;
            var width = Math.Abs(Hook.StartPoint.X - point.X);
            var height = Math.Abs(Hook.StartPoint.Y - point.Y);
            var x = Hook.StartPoint.X <= point.X ? Hook.StartPoint.X : point.X;
            var y = Hook.StartPoint.Y <= point.Y ? Hook.StartPoint.Y : point.Y;
            var screenPoint = new Point(x, y);
            var componentInfo = VisToolboxItem.SelectedTool.ComponentInfo;

            if (SelectionBox.DrawingBox != null) {
                SelectionBox.DrawingBox.Close(); SelectionBox.DrawingBox = null;
            }
            if (VisToolboxItem.IsOneSelected)
                VisToolboxItem.Deselect();
            Hook.State = new HookNeutralState();  // The change command below may generate MsgBoxes. Allow simple cursor move.
            Point relativePoint;
            container = FindContainer(container, screenPoint);
            if (container == null)
                relativePoint = screenPoint;
            else 
                relativePoint = container.GuiInstance.PointToClient(screenPoint);
            controller.TemplateDrawn(componentInfo, container, relativePoint, width, height);
        }
    }

    //----------------------------------------------------------------------------------------
    public class MoveTransaction {
        public VisComponent Instance;
        //public string Container;
        public Template tpl;
        public FormTemplate newFrmTpl;
        public int NewLeft; // Logical position
        public int NewTop; // Logical position
        public string CanvasFormula = string.Empty;
        // Some components implement position properties through other properties, e.g. Polygon uses Xpositions. 
        // This requires saving and restoring all value property formulas. 
        public string[] oldFormulas;

        public MoveTransaction(VisComponent instance, FormTemplate newFrmTpl) { // newForm null if move inside the form.
            this.Instance = instance;
            if (instance == null) return; // Used by templatesCopied, which sets tpl itself. 
            this.tpl = instance.tpl;
            this.newFrmTpl = newFrmTpl;

            CanvasFormula = tpl.propCanvas.Formula;
            int valLength = tpl.valPropArray.Length;
            oldFormulas = new string[valLength];
            for (int i = 0; i < valLength; i++)
                oldFormulas[i] = tpl.valPropArray[i].Formula;
        }
    }

    //----------------------------------------------------------------------------------------
    public class ResizeTransaction {
        public VisComponent Instance;
        public Template tpl;
        public int BeforeLeft; // Logical positions and sizes.
        public int BeforeTop;
        public int BeforeWidth;
        public int BeforeHeight;
        public int NewLeft;
        public int NewTop;
        public int NewWidth;
        public int NewHeight;
        // Some components implement position properties through other properties, e.g. Polygon uses Xpositions. 
        // This requires saving and restoring all value property formulas. 
        public string[] oldFormulas;

        public ResizeTransaction(VisComponent instance) {
            this.Instance = instance;
            this.tpl = instance.tpl;

            this.BeforeLeft = instance.Left;
            this.BeforeTop = instance.Top;
            this.BeforeWidth = instance.Width;
            this.BeforeHeight = instance.Height;

            int valLength = tpl.valPropArray.Length;
            oldFormulas = new string[valLength];
            for (int i = 0; i < valLength; i++)
                oldFormulas[i] = tpl.valPropArray[i].Formula;
        }
    }
}
