﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using VisTool;

namespace uVisStudio.VisDesigner {
    public enum HookResizeMode {
        Nothing,
        West,
        East,
        North,
        South
    }

    //----------------------------------------------------------
    public class BlueBox : Form { // A transparent form without border.
        // Mouse responds only on the border. Must be moved programmatically. We use MouseCapture to move it across the screen - other ways?
        // TransparencyKey works only when the Form is TopLevel. But then it doesn't move automatically with the form 
        // and doesn't close automatically.
        const int BorderWidth = 2;

        public BlueBox(Point screenLoc, Size size) : base() {
            this.ShowInTaskbar = false;
            this.ControlBox = false;
            this.FormBorderStyle = FormBorderStyle.None;
            this.TopMost = true; // Above normal forms.
            // this.TopLevel = false; // If false, TransparencyKey doesn't work.
            this.StartPosition = FormStartPosition.Manual;

            int Htrim = size.Width > 10 ? 0 : 5 - size.Width / 2; // Make sure that large components such as canvas don't steal neigbor areas.
            int Vtrim = size.Height > 10 ? 0 : 5 - size.Height / 2;
            this.Left = screenLoc.X - Htrim; // Set location before setting size.
            this.Top = screenLoc.Y - Vtrim; 
            this.TransparencyKey = this.BackColor; // Pixels with this color will not be painted.
            this.Show(); // For strange reasons, Show has to be here. If it is after setting Width and Height, they become wrong.
            this.Width = size.Width + 2 * Htrim;
            this.Height = size.Height + 2 * Vtrim;
            this.Paint += OnPaint;
        }

        public void OnPaint(object sender, PaintEventArgs e) {
            e.Graphics.SmoothingMode = SmoothingMode.HighSpeed;
            Pen pen = new Pen(Color.CornflowerBlue, BorderWidth);
            pen.DashStyle = DashStyle.Dash;
            e.Graphics.DrawRectangle(pen, BorderWidth/2, BorderWidth/2, this.Width - BorderWidth, this.Height - BorderWidth);
        }
    }

    //-----------------------------------------------------------------
    public class SelectionBox { // Data about a design-selected component, e.g. the Blue box that shows it.
        const int weight = 3; // Half width of the sensitive area.

        public BlueBox blueBox;
        public VisComponent cmp;
        public Template tpl;
        public VisForm frm;
        public Form form;
        public int relativeX, relativeY; // Position of blue box relative to form. Used for dragging. (4, 23) for a Form.

        // Static data: Dictionary of SelectionBoxes and data about the currently selected one.
        public static Dictionary<Template, SelectionBox> selections = new Dictionary<Template, SelectionBox>(10);
        // Only one component of a template can be selected.
        public static SelectionBox selected = null; // Refers to one of the selections. Null when selections.Count == 0.
        public static BlueBox DrawingBox; 

        public static int dragOffsetX = 0; // Distance from top-left corner to cursor when dragging.
        public static int dragOffsetY = 0;
        public static List<VisComponent> clickedStack = new List<VisComponent>(); // The stack of components under the cursor.
        public static int latestIndex = 0; // Index in the clicked stack. To select the next in the stack.

        public SelectionBox(VisComponent cmp) { // Creates a selection, adds it to selections and lets selected refer to it.
            this.cmp = cmp;
            this.tpl = cmp.Bundle.template;
            DeleteBox(tpl); // Only one component per template can be selected. Delete a possibly old one.
            VisComponent c = cmp;
            while (!(c is VisForm)) c = c.Bundle.Parent;
            frm = c as VisForm;
            form = frm.form;
            Size size = cmp is VisForm ? form.ClientSize : new Size(cmp.width, cmp.height);
            Point topLeft = cmp.PointToScreen(new Point(0, 0)); // For a Form it is topLeft of the client area.
            this.relativeX = topLeft.X - form.Left;
            this.relativeY = topLeft.Y - form.Top; // (4, 23) for VisForm.
            blueBox = new BlueBox(topLeft, size);
            selections.Add(tpl, this);
            selected = this;
            //form.Controls.Add(blueBox); When blueBox is a child of the main form, it cannot be transparent.
        }

        public static void DeleteBox(Template tpl) {
            SelectionBox del;
            if (selections.TryGetValue(tpl, out del)) {
                del.blueBox.Close();
                selections.Remove(tpl);
                if (selections.Count == 0)
                    selected = null;
                else {
                    foreach (var kv in selections) { selected = kv.Value; break; }
                }
            }
        }

        public static void DeleteAll() { // Delete all boxes.
            foreach (var kv in selections) {
                kv.Value.blueBox.Close(); // Don't use DeleteBox. Disturbs the iterator.
            }
            selections.Clear(); selected = null;
        }

        public static void ResetAll() { // Reset all selection boxes to the unmoved components.
            foreach (var kv in selections) {
                SelectionBox sel = kv.Value;
                VisComponent cmp = sel.cmp;
                Size size = cmp is VisForm ? sel.form.ClientRectangle.Size : new Size(cmp.width, cmp.height); 
                Point topLeft = cmp.PointToScreen(new Point(0, 0));
                sel.blueBox.Location = topLeft;
                sel.blueBox.Size = size;
                sel.blueBox.Invalidate(); // Otherwise old blue box may be left on screen.
            }
        }

        public static SelectionBox SelectionClicked(Point point) { // Find any selected component that contains point.
            foreach (var kv in selections) {
                SelectionBox sel = kv.Value;
                VisComponent cmp = sel.cmp;
                Size size = cmp is VisForm ? sel.form.ClientRectangle.Size : new Size(cmp.width, cmp.height);
                Point topLeft = cmp.PointToScreen(new Point(0, 0));
                var area = new Rectangle(topLeft.X - weight, topLeft.Y - weight, size.Width + weight * 2, size.Height + weight * 2);
                if (area.Contains(point)) return kv.Value;
            }
            return null;
        }

        public HookResizeMode ResizeMode(Point point) { 
            // Checks whether point in screen coordinates is on a border of the selected component. 
            BlueBox box = this.blueBox; // Box may be slightly larger than cmp.
            Point topLeft = box.Location; // PointToScreen(new Point(0, 0));
            // For VisForms, the top-left corner of the client area.
            int width = this.cmp is VisForm ? this.form.ClientRectangle.Width : box.Width;
            int height = this.cmp is VisForm ? this.form.ClientRectangle.Height : box.Height;

            var west = new Rectangle(topLeft.X - weight, topLeft.Y, weight+1, height);
            var north = new Rectangle(topLeft.X, topLeft.Y - weight, width, weight+1);
            var east = new Rectangle(topLeft.X + width - weight, topLeft.Y, weight +1, height);
            var south = new Rectangle(topLeft.X, topLeft.Y + height - weight, width, weight +1);

            if (west.Contains(point)) return HookResizeMode.West;
            if (north.Contains(point)) return HookResizeMode.North;
            if (east.Contains(point)) return HookResizeMode.East;
            if (south.Contains(point)) return HookResizeMode.South;
            return HookResizeMode.Nothing;
        }

        public bool IsMoveable(Point point) { // Checks whether point in screen coordinates is inside the selected component. 
            BlueBox box = this.blueBox; // Box may be slightly larger than cmp.
            Point topLeft = box.Location; // PointToScreen(new Point(0, 0));
            Size size = this.cmp is VisForm ? this.form.ClientRectangle.Size : new Size(box.Width, box.Height);
            var inner = new Rectangle(topLeft.X + 1, topLeft.Y, box.Width, box.Height);
            return inner.Contains(point);
        }
    }
}
